from typing import *

import attr
from asn1crypto.core import Choice, Integer

from dlms_cosem import enumerations
from dlms_cosem.ber import BER
from dlms_cosem.protocol.acse import base as acse_base
from dlms_cosem.protocol.acse.aarq import aarq_should_set_authenticated
from dlms_cosem.protocol.acse.user_information import UserInformation


@attr.s(auto_attribs=True)
class Asn1Integer:

    """
    Simple class to wrap Integers for BER encoding.
    Does not handle integers larger than 128 yet
    """

    value: int

    TAG: ClassVar[int] = 2  # ASN1 Universal tag 2, Integer.

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        tag, length, data = BER.decode(source_bytes)
        if tag != cls.TAG.to_bytes(1, "big"):
            raise ValueError(
                f"Data provided is not of the correct type. Tag is {tag} but should "
                f"be {cls.TAG}"
            )
        return cls(value=int.from_bytes(data, byteorder="big"))

    def to_bytes(self):
        return BER.encode(self.TAG, self.value.to_bytes(1, byteorder="big"))


class ResultSourceDiagnostics(Choice):
    _alternatives = [
        ("acse-service-user", Integer, {"explicit": 1}),
        ("acse-service-provider", Integer, {"explicit": 2}),
    ]

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        return cls.load(source_bytes)


@attr.s(auto_attribs=True)
class ApplicationAssociationResponse(acse_base.AbstractAcseApdu):
    """
    AARE-apdu ::= [APPLICATION 1] IMPLICIT SEQUENCE
        APPLICATION 1 = 0x61 = 97

        protocol-version            [0] IMPLICIT    BIT STRING {version1 (0)} DEFAULT {version1},
        application-context-name    [1]             Application-context-name,
        result                      [2]             Association-result,
        result-source-diagnostic    [3]             Associate-source-diagnostic,
        responding-AP-title         [4]             AP-title OPTIONAL,
        responding-AE-qualifier     [5]             AE-qualifier OPTIONAL,
        responding-AP-invocation-id [6]             AP-invocation-identifier OPTIONAL,
        responding-AE-invocation-id [7] AE-invocation-identifier OPTIONAL,

        -- The following field shall not be present if only the kernel is used.
        responder-acse-requirements [8] IMPLICIT    ACSE-requirements OPTIONAL,

        -- The following field shall only be present if the authentication functional unit is selected.
        mechanism-name [9] IMPLICIT                 Mechanism-name OPTIONAL,

        -- The following field shall only be present if the authentication functional unit is selected.
        responding-authentication-value [10] EXPLICIT   Authentication-value OPTIONAL,
        implementation-information [29] IMPLICIT        Implementation-data OPTIONAL,
        user-information            [30] EXPLICIT       Association-information OPTIONAL

        -- The user-information field shall carry either an InitiateResponse
            (or, when the proposed xDLMS -- context is not accepted by the server,
            a confirmedServiceError) APDU encoded in A-XDR, and then
            -- encoding the resulting OCTET STRING in BER.

    :parameter result: In the case of remote confirmation in shows if the server (meter)
        accepted the application association request. In the case of local confirmation
        in shows if the client accepted

    # TODO: Exactly what is the difference between remot and local confirmation.
    # My guess is that it is always remote for this library's usage.

     :parameter meter_system_title:  It is transferred in the `responding_ap_title`
        portion from the DLMS ASN1 specs.
        If the negotiated `application_context_name` uses
        ciphering `responding_ap_title` should contain the server system title. Also if
        the negotiated HLS (High Level Security) auth mechanism uses the server system
        title it should be present.

    :parameter meter_public_cert: Is transferred in the `responding_ae_qualifier` of the
        ASN1 DLMS Spec `responding_ae_qualifier`. If the `application_context_name`
        indicates the
        use of ciphered APDUs the `responding_ae_qualifier` may hold the public digital
        signature key certificate of the server (meter).

    :parameter authentication_value: Is transferred in the
        `responding_authentication_value` field from the ASN1 DLMS specs. Hold the
        server (meter) challenge to the current authentication scheme.

    """

    TAG: ClassVar[int] = 0x61

    PARSE_TAGS = {
        128: ("protocol_version", None),  # Context specific, constructed? 0
        161: ("application_context_name", acse_base.AppContextName),
        # Context specific, constructed 1
        162: ("result", Asn1Integer),
        163: ("result_source_diagnostics", ResultSourceDiagnostics),
        164: ("responding_ap_title", None),
        165: ("responding_ae_qualifier", None),
        166: ("responding_ap_invocation_id", None),
        167: ("responding_ae_invocation_id", None),
        0x88: ("responder_acse_requirements", None),
        0x89: ("mechanism_name", acse_base.MechanismName),
        170: ("responding_authentication_value", acse_base.AuthenticationValue),
        189: ("implementation_information", None),
        0xBE: ("user_information", UserInformation),  # Context specific, constructed 30
    }

    result: enumerations.AssociationResult
    result_source_diagnostics: Union[
        enumerations.AcseServiceUserDiagnostics,
        enumerations.AcseServiceProviderDiagnostics,
    ]
    ciphered: bool = attr.ib(default=False)
    authentication: Optional[enumerations.AuthenticationMechanism] = attr.ib(
        default=None
    )
    system_title: Optional[bytes] = attr.ib(default=None)
    public_cert: Optional[bytes] = attr.ib(default=None)
    authentication_value: Optional[bytes] = attr.ib(default=None)
    user_information: Optional[UserInformation] = attr.ib(default=None)

    # Not really used.
    implementation_information: Optional[bytes] = attr.ib(default=None)
    responding_ap_invocation_id: Optional[bytes] = attr.ib(default=None)  # Not used.
    responding_ae_invocation_id: Optional[bytes] = attr.ib(default=None)  # Not used.

    @property
    def responder_acse_requirements(self) -> Optional[acse_base.AuthFunctionalUnit]:
        """
        Is only sent if the AuthFunctionalUnit needs to indicate authentication.
        """
        if aarq_should_set_authenticated(self.authentication):
            return acse_base.AuthFunctionalUnit(True)

        return None

    @property
    def mechanism_name(self) -> Optional[acse_base.MechanismName]:
        """
        The mechanism_name field should only be present if the AuthFunctionalUnit
        indicates authenticated.
        :return:
        """
        if (
            self.responder_acse_requirements is not None
            and self.authentication is not None
        ):
            return acse_base.MechanismName(mechanism=self.authentication)

        return None

    @property
    def application_context_name(self) -> acse_base.AppContextName:
        """
        Always use logical name references.
        """
        if self.ciphered:
            return acse_base.AppContextName(logical_name_refs=True, ciphered_apdus=True)
        else:
            return acse_base.AppContextName(
                logical_name_refs=True, ciphered_apdus=False
            )

    @property
    def protocol_version(self) -> int:
        return 0

    @classmethod
    def from_bytes(cls, source_bytes: bytes):

        aare_data = bytearray(source_bytes)

        aare_tag = aare_data.pop(0)
        if not aare_tag == cls.TAG:
            raise ValueError("Bytes are not an AARQ APDU. TAg is not int(96)")

        aare_length = aare_data.pop(0)

        if not len(aare_data) == aare_length:
            raise ValueError(
                "The APDU Data lenght does not correspond " "to length byte"
            )

        # Assumes that the protocol-version is 1 and we don't need to decode it

        # Decode the AARQ  data
        object_dict = dict()

        # use the data in tags to go through the bytes and create objects.
        while True:
            # TODO: this does not take into account when defining objects in dict and not using them.
            object_tag = aare_data.pop(0)
            object_desc = ApplicationAssociationResponse.PARSE_TAGS.get(
                object_tag, None
            )
            if object_desc is None:
                raise ValueError(
                    f"Could not find object with tag {object_tag} "
                    f"in AARQ definition"
                )
            object_length = aare_data.pop(0)
            object_data = bytes(aare_data[:object_length])
            aare_data = aare_data[object_length:]

            object_name = object_desc[0]
            object_class: Any = object_desc[1]

            if object_class is not None:
                object_data = object_class.from_bytes(object_data)

            object_dict[object_name] = object_data

            if len(aare_data) <= 0:
                break

        context_name: acse_base.AppContextName = object_dict.pop(
            "application_context_name"
        )
        object_dict["ciphered"] = context_name.ciphered_apdus

        if not context_name.logical_name_refs:
            raise ValueError(
                "AARE requests use of Short Name referencing which is not " "supported."
            )

        protocol_version: Optional[int] = object_dict.pop("protocol_version", None)
        if protocol_version is not None:
            if protocol_version != 0:
                raise ValueError("Parsed a protocol version that is not 0")

        # transform the result into an enum
        object_dict["result"] = enumerations.AssociationResult(
            object_dict["result"].value
        )

        # tarnsorm the source diagnositc into enum
        source_diagnostic = object_dict["result_source_diagnostics"]

        if source_diagnostic.name == "acse-service-user":
            object_dict[
                "result_source_diagnostics"
            ] = enumerations.AcseServiceUserDiagnostics(source_diagnostic.native)
        elif source_diagnostic.name == "acse-service-provider":
            object_dict[
                "result_source_diagnostics"
            ] = enumerations.AcseServiceProviderDiagnostics(source_diagnostic.native)

        else:
            raise ValueError("Not a valid choice of result_source_diagnostics")

        responder_acse_requirements: Optional[
            acse_base.AuthFunctionalUnit
        ] = object_dict.pop("responder_acse_requirements", None)

        mechanism_name: Optional[acse_base.MechanismName] = object_dict.pop(
            "mechanism_name", None
        )

        if responder_acse_requirements and mechanism_name:
            object_dict["authentication"] = mechanism_name.mechanism
        # rename responding_ap_title to meter_system_title for cleaner API
        # And it is ber encoded octet string universal 4. quick and diryt parse.
        meter_system_title = object_dict.pop("responding_ap_title", None)
        if meter_system_title:
            # it is ber encoded universal tag ocetctring. simple handling
            object_dict["system_title"] = bytes(meter_system_title[2:])
        else:
            object_dict["system_title"] = None
        # rename responding_ae_qualifier to meter_public_cert
        meter_public_cert = object_dict.pop("responding_ae_qualifier", None)
        if meter_public_cert:
            # it is ber encoded universal tag ocetctring. simple handling
            object_dict["public_cert"] = bytes(meter_public_cert[2:])
        else:
            object_dict["public_cert"] = None

        auth_value: Optional[acse_base.AuthenticationValue] = object_dict.pop(
            "responding_authentication_value", None
        )
        if auth_value:
            object_dict["authentication_value"] = bytes(auth_value.password)
        else:
            object_dict["authentication_value"] = None

        return cls(**object_dict)

    def to_bytes(self) -> bytes:

        aare_data = bytearray()
        # default value of protocol_version is 1. Only decode if other than 1
        # No need to use protocol version
        # if self.protocol_version != 1:
        #    aare_data.extend(BER.encode(160, bytes(self.protocol_version)))
        if self.application_context_name is not None:
            aare_data.extend(BER.encode(161, self.application_context_name.to_bytes()))
        if self.result is not None:
            aare_data.extend(
                BER.encode(162, Asn1Integer(value=self.result.value).to_bytes())
            )
        if self.result_source_diagnostics is not None:
            if isinstance(
                self.result_source_diagnostics, enumerations.AcseServiceUserDiagnostics
            ):
                aare_data.extend(
                    BER.encode(
                        163,
                        ResultSourceDiagnostics(
                            name="acse-service-user",
                            value=self.result_source_diagnostics.value,
                        ).dump(),
                    )
                )
            elif isinstance(
                self.result_source_diagnostics,
                enumerations.AcseServiceProviderDiagnostics,
            ):
                aare_data.extend(
                    BER.encode(
                        163,
                        ResultSourceDiagnostics(
                            name="acse-service-provider",
                            value=self.result_source_diagnostics.value,
                        ).dump(),
                    )
                )

        if self.system_title is not None:
            aare_data.extend(BER.encode(164, BER.encode(4, self.system_title)))
        if self.public_cert is not None:
            aare_data.extend(BER.encode(165, BER.encode(4, self.public_cert)))
        if self.responding_ap_invocation_id is not None:
            aare_data.extend(BER.encode(166, self.responding_ap_invocation_id))
        if self.responding_ae_invocation_id is not None:
            aare_data.extend(BER.encode(167, self.responding_ae_invocation_id))
        if self.responder_acse_requirements is not None:
            aare_data.extend(
                BER.encode(0x88, self.responder_acse_requirements.to_bytes())
            )
        if self.mechanism_name is not None:
            aare_data.extend(BER.encode(0x89, self.mechanism_name.to_bytes()))
        if self.authentication_value is not None:
            aare_data.extend(
                BER.encode(
                    170,
                    acse_base.AuthenticationValue(
                        password=self.authentication_value
                    ).to_bytes(),
                )
            )

        if self.implementation_information is not None:
            aare_data.extend(BER.encode(189, self.implementation_information))
        if self.user_information is not None:
            aare_data.extend(BER.encode(0xBE, self.user_information.to_bytes()))

        return BER.encode(self.TAG, aare_data)



import datetime

import pytest
from dateutil.parser import parse as dt_parse

from dlms_cosem.time import datetime_from_bytes, datetime_to_bytes


@pytest.mark.parametrize(
    "bytes_representation, dt",
    [
        (
            b"\x07\xe4\x01\x01\xff\x00\x03\x00\x00\xff\x88\x00",
            dt_parse("2020-01-01T00:03:00+02:00"),
        ),
        (
            b"\x07\xe4\x01\x06\xff\x00\x03\x00\x00\xff\xc4\x00",
            dt_parse("2020-01-06T00:03:00+01:00"),
        ),
        (
            b"\x07\xe2\x02\x0c\xff\x00\x00\x00\x00\x80\x00\x00",
            dt_parse("2018-02-12T00:00:00"),
        ),
    ],
)
def test_bytes_datetime_conversion(bytes_representation: bytes, dt: datetime.datetime):

    assert datetime_from_bytes(bytes_representation)[0] == dt
    assert datetime_to_bytes(dt) == bytes_representation



import pytest

from dlms_cosem import utils
from dlms_cosem.protocol import xdlms


class TestDataNotification:
    def test_transform_bytes(self):
        dlms_data = b'\x0f\x00\x00\x01\xdb\x00\t"\x12Z\x85\x916\x00\x00\x00\x00I\x00\x00\x00\x11\x00\x00\x00\nZ\x85\x13\xd0\x14\x80\x00\x00\x00\r\x00\x00\x00\n\x01\x00'
        data_notification = xdlms.DataNotification.from_bytes(dlms_data)
        print(data_notification)
        assert data_notification.date_time is None
        assert (
            data_notification.body
            == b'\t"\x12Z\x85\x916\x00\x00\x00\x00I\x00\x00\x00\x11\x00\x00\x00\nZ\x85\x13\xd0\x14\x80\x00\x00\x00\r\x00\x00\x00\n\x01\x00'
        )
        utils.parse_as_dlms_data(data_notification.body)
        assert data_notification.to_bytes() == dlms_data



from typing import *

import attr

# TODO: when using ciphered apdus we will get other apdus. (33 64) global or dedicated cipered iniitate requests


@attr.s(auto_attribs=True)
class Conformance:
    """
    Holds information about the supported services in a DLMS association.
    Is used to send the propsed conformance in AARQ and to send back the negotiated
    conformance in the AARE.

    Only LN referenceing is supported.

    It specifes Conformance ::= [Application 31] Implicit BIT STRING (24)
    Conformance should be BER encoded.
    Having tags that are higher than 31 in BER encoding gives multi byte tags.
    Where the first tag byte fills all tag values. And the consecutive bytes
    represents the tag number. If the number cannot be contined in one extra byte the
    msb should be set to 1 and bytes added with msb set to 1 until the tag number can
    fit. The last byte should have the msb set to 0

    So for Application 31 the tag is 0x01011111 + 0b00011111 = 0x5f 0x1F.

    # TODO: how to code than.....

    But in the example they encode it {0x5F 0x1F {length=0x04} {number of unused bits last byte} {3x byte for bitstring(24}}
    The ASN.1 shoudl give 0x5

    The bit placement for the conformance flags are also a bit weird in the standard.
    Since they count the bits from left to right. So bit 0 is the MSB.
    The placement in this class sets them as bit 0 is LSB.

    # TODO: Should also be set up at the assosiation to track the negotiated.
    """

    general_protection: bool = attr.ib(default=False)
    general_block_transfer: bool = attr.ib(default=False)
    delta_value_encoding: bool = attr.ib(default=False)
    attribute_0_supported_with_set: bool = attr.ib(default=False)
    priority_management_supported: bool = attr.ib(default=False)
    attribute_0_supported_with_get: bool = attr.ib(default=False)
    block_transfer_with_get_or_read: bool = attr.ib(default=False)
    block_transfer_with_set_or_write: bool = attr.ib(default=False)
    block_transfer_with_action: bool = attr.ib(default=False)
    multiple_references: bool = attr.ib(default=False)
    data_notification: bool = attr.ib(default=False)
    access: bool = attr.ib(default=False)
    get: bool = attr.ib(default=False)
    set: bool = attr.ib(default=False)
    selective_access: bool = attr.ib(default=False)
    event_notification: bool = attr.ib(default=False)
    action: bool = attr.ib(default=False)

    # bit numbering starts at 0
    conformance_bit_position: ClassVar[Dict[str, int]] = {
        "general_protection": 22,
        "general_block_transfer": 21,
        "delta_value_encoding": 17,
        "attribute_0_supported_with_set": 15,
        "priority_management_supported": 14,
        "attribute_0_supported_with_get": 13,
        "block_transfer_with_get_or_read": 12,
        "block_transfer_with_set_or_write": 11,
        "block_transfer_with_action": 10,
        "multiple_references": 9,
        "data_notification": 7,
        "access": 6,
        "get": 4,
        "set": 3,
        "selective_access": 2,
        "event_notification": 1,
        "action": 0,
    }

    @classmethod
    def from_bytes(cls, in_bytes: bytes):
        in_dict = dict()
        integer_representation = int.from_bytes(in_bytes[1:], "big")
        for attribute in Conformance.conformance_bit_position.keys():
            in_dict[attribute] = bool(
                integer_representation
                & (1 << Conformance.conformance_bit_position[attribute])
            )
        return cls(**in_dict)

    def to_bytes(self):
        out = 0
        for attribute, position in Conformance.conformance_bit_position.items():
            flag_is_set = getattr(self, attribute)
            if flag_is_set:
                out += 1 << position
        # It is a bit string so need to encode how many bits that are unused in the
        # last byte. Its none so we can just put 0x00 infront.
        return b"\x00" + out.to_bytes(3, "big")



"""
A XDR (IEC 61334-6is an Adapted External Data Represerntation Standard
(XDR for DLMS)

It usage is for the reduction of APDU sizes and saving bandwidth by eliminating
data that is already known to both sender and receiver.

For example in comparison to BER encoding where the length of data is encoded
using A-XDR the lenght byte could be ommited if both sender and receiver are
aware of the lenght of an integer. (2 bytes for example.)

It is used to encode xDLMS APDUs.

It is not used for AARQ, AARE, RLRQ, and RLRE. (BER is used)

.. note:
    The InititiateRequest the above requests and responses is an xDLMS APDU and
    uses A-XDR.


Optional Values can be omitted by encoding 0x00 in its place. If a value is used
it should be preceded with 0x01. (0x01+data)

Default values are encoded with 0x00 if using the default and 0x01+data if
using non default.

Encoding integers: A-XDR makes it possible to encode with a fixed range and a
variable range.

Fixed range integers are encoded with the minimum number of bytes needed to fit
the value range.

Variable range integers use the leftmost bit to control the encoding.
leftmost bit = 0. Value < 128 , value can be encoded in one byte
leftmost bit = 1. The whole leftmost byte is used to indicate the lenght of the
integer data. ex 0b10000010 -> 2 bytes after this is the integer. 0x820xff0xff = 65535


"""

from typing import *

import attr

from dlms_cosem import dlms_data

VARIABLE_LENGTH = -1


def get_axdr_length(data: bytearray):
    """
    Will find the length of an xadr element assuming the length is the first bytes in
    the data
    Works with bytearray and will remove element from the array as it finds the
    variable length.
    """
    length_data = bytearray()
    first_byte = data.pop(0)
    length_is_multiple_bytes = bool(first_byte & 0b10000000)
    if not length_is_multiple_bytes:
        return first_byte
    number_of_bytes_representing_the_length = first_byte & 0b01111111
    for _ in range(0, number_of_bytes_representing_the_length):
        length_data.append(data.pop(0))
    return int.from_bytes(length_data, "big")


@attr.s
class DataSequenceEncoding:
    attribute_name: str = attr.ib()


@attr.s(auto_attribs=True)
class Attribute:
    attribute_name: str
    create_instance: Callable
    length: int = attr.ib(default=VARIABLE_LENGTH)
    return_value: Optional[bool] = attr.ib(default=False)
    wrap_end: Optional[bool] = attr.ib(default=False)  # Maybe name wrapper?
    default: Optional[Any] = attr.ib(default=None)
    optional: Optional[bool] = attr.ib(default=False)


@attr.s(auto_attribs=True)
class Sequence:

    attribute_name: str
    instance_factory: dlms_data.DlmsDataFactory = attr.ib(
        factory=dlms_data.DlmsDataFactory
    )


@attr.s(auto_attribs=True)
class Choice:
    choices: Dict[bytes, Union[Attribute, Sequence]]


@attr.s(auto_attribs=True)
class EncodingConf:
    attributes: List[Union[Attribute, Sequence, Choice]]


# TODO: we need to be able to fix the lenght of variable lenght data.
# TODO: if it is the last element give it all data left.


@attr.s(auto_attribs=True)
class AXdrDecoder:
    encoding_conf: EncodingConf
    buffer: bytearray = attr.ib(factory=bytearray)
    pointer: int = attr.ib(default=0)
    result: Dict[str, Any] = attr.ib(factory=list)

    @property
    def buffer_empty(self) -> bool:
        return self.pointer == len(self.buffer)

    def decode(self, data: bytes):
        # clear previous results
        self.result = dict()
        # fill the buffer
        self.buffer += data
        for index, data_attribute in enumerate(self.encoding_conf.attributes):
            self.result.update(self.decode_single(data_attribute, index))

        return self.result

    def is_last_encoding_element(self, index: int) -> bool:
        return index == len(self.encoding_conf.attributes)

    def get_buffer_tail(self) -> bytearray:
        return self.buffer[self.pointer :]

    def decode_single(self, _type, index: int) -> Dict:
        if isinstance(_type, Attribute):
            return {_type.attribute_name: self.decode_attribute(_type, index)}

        elif isinstance(_type, Choice):
            choice = _type.choices[bytes(self.get_bytes(1))]
            return self.decode_single(choice, index)

        elif isinstance(_type, Sequence):
            return self.decode_sequence(_type)
        else:
            raise ValueError("No valid class type")

    def decode_attribute(self, attribute: Attribute, index: int) -> Optional[Any]:
        if attribute.optional:
            indicator = self.get_bytes(1)
            if indicator == b"\x00":
                # Not used.
                return None

        if attribute.default is not None:
            indicator = self.get_bytes(1)
            if indicator == b"\x00":
                # use the default
                return attribute.default

        # fixed lenght?  # TODO: Is all attributes of fixed lenght in X-ADR?
        if attribute.length != VARIABLE_LENGTH:
            data = self.get_bytes(attribute.length)
            return attribute.create_instance(data)
        else:
            # check if last element
            if self.is_last_encoding_element(index):
                return attribute.create_instance()
            # We know hot to create the instance (just not how long it is)
            length = self.get_axdr_length()
            data = self.get_bytes(length)
            return attribute.create_instance(data)

    def decode_fixed_length_attribute(self, encoding: Attribute, data) -> Any:
        """
        When we know the encoding of a fixed length value we can just feed the data to
        the instance_creator
        """
        return encoding.create_instance(data)

    def decode_sequence(self, seq: Sequence) -> Dict:
        parsed_data = list()

        while not self.buffer_empty:
            tag = self.get_bytes(1)

            data_class = dlms_data.DlmsDataFactory.get_data_class(
                int.from_bytes(tag, "big")
            )

            if data_class == dlms_data.DataArray:
                parsed_data.append(self.decode_array())
                continue

            if data_class == dlms_data.DataStructure:
                parsed_data.append(self.decode_structure())
                continue

            if data_class.LENGTH != VARIABLE_LENGTH:
                parsed_data.append(
                    data_class.from_bytes(
                        bytes(self.get_bytes(data_class.LENGTH))
                    ).to_python()
                )
                continue

            # TODO: should have a function to get variable intefer incase it is longer
            #   than what a normal byte can handle.

            length_or_items = self.get_axdr_length()
            parsed_data.append(
                data_class.from_bytes(
                    bytes(self.get_bytes(length_or_items))
                ).to_python()
            )
            continue

        if len(parsed_data) == 1:
            return {seq.attribute_name: parsed_data[0]}

        return {seq.attribute_name: parsed_data}

    def decode_sequence_of(self):

        tag = int.from_bytes(self.get_bytes(1), "big")
        data_class = dlms_data.DlmsDataFactory.get_data_class(tag)

        if data_class == dlms_data.DataArray:
            return self.decode_array()

        if data_class == dlms_data.DataStructure:
            return self.decode_structure()

        else:
            return self.decode_data(data_class)

    def decode_data(self, data_class):
        assert data_class not in [dlms_data.DataArray, dlms_data.DataStructure]

        if data_class.LENGTH == VARIABLE_LENGTH:
            length = self.get_axdr_length()
            return data_class.from_bytes(self.get_bytes(length)).to_python()
        else:
            return data_class.from_bytes(self.get_bytes(data_class.LENGTH)).to_python()

    def decode_array(self):
        item_count = self.get_axdr_length()
        elements = list()
        for _ in range(0, item_count):
            elements.append(self.decode_sequence_of())
        return elements

    def decode_structure(self):
        item_count = self.get_axdr_length()
        elements = list()
        for _ in range(0, item_count):
            elements.append(self.decode_sequence_of())

        return elements

    def get_bytes(self, length: int) -> bytearray:
        """Gets some bytes from the buffer and moves the pointer forward."""
        part = self.buffer[self.pointer : self.pointer + length]
        self.pointer += length
        return part

    @property
    def remaining_buffer(self) -> bytearray:
        return self.buffer[self.pointer :]

    def get_axdr_length(self) -> int:
        length_data = bytearray()
        first_byte = int.from_bytes(self.get_bytes(1), "big")
        length_is_multiple_bytes = bool(first_byte & 0b10000000)
        if not length_is_multiple_bytes:
            return first_byte
        number_of_bytes_representing_the_length = first_byte & 0b01111111
        for _ in range(0, number_of_bytes_representing_the_length):
            length_data.extend(self.get_bytes(1))
        return int.from_bytes(length_data, "big")


class DlmsDataToPythonConverter:
    def __init__(self, encoding_conf: List[dlms_data.BaseDlmsData]):
        self.encoding_conf = encoding_conf

    def to_python(self):
        out_list = list()
        for item in self.encoding_conf:
            out_list.append(item.value)

        return out_list

    def to_dlms(self, data: List):
        raise NotImplementedError("Not yet supported to convert python values to DLMS")



from __future__ import annotations  # noqa

import os
from typing import Optional, ClassVar


import attr
from cryptography.exceptions import InvalidTag
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.keywrap import aes_key_unwrap, aes_key_wrap

from dlms_cosem import enumerations, exceptions

from dlms_cosem.exceptions import CipheringError, DecryptionError


import sys
from typing import TYPE_CHECKING

if sys.version_info < (3, 8):
    from typing_extensions import Protocol
else:
    from typing import Protocol

if TYPE_CHECKING:
    from dlms_cosem.connection import DlmsConnection, ProtectionError


"""
Security Suites in DLMS/COSEM define what cryptographic algorithms that are
available to different services and key sizes

The initialization vector is essentially a nonce. In DLMS/COSEM it is
    composed of two parts. The full length is 96 bits (12 bytes)
    The first part (upper 64bit/8bytes) is called the fixed field and shall
    contain the system title. The lower (32bit/4byte) part is called the
    invocation field and contains an integer invocation counter.
    The system title is a unique identifier for the DLMS/COSEM identity. The
    leftmost 3 octets holds the 3 letter manufacturer ID. (FLAG ID) and the
    remaining 5 octets are to ensure uniqueness.
"""

TAG_LENGTH = 12


def validate_security_suite_number(instance, attribute, value):
    if value not in [0, 1, 2]:
        raise ValueError(f"Only Security Suite 0-2 is valid, Got: {value}")


@attr.s(auto_attribs=True)
class SecurityControlField:
    """
    8 bit unsigned integer

    Bit 3...0: Security Suite number
    Bit 4: Indicates if authentication is applied
    Bit 5: Indicates if encryption is applied
    Bit 6: Key usage: 0 = Unicast Encryption Key , 1 = Broadcast Encryption Key
    Bit 7: Indicates the use of compression

    :param bool security_suite: Number of the DLMS Security Suite used, valid
        are 1, 2, 3.
    :param bool authenticated: Indicates if authentication is applied
    :param bool encrypted: Indicates if encryption is applied
    :param bool broadcast_key: Indicates use of broadcast key. If false unicast key is used.
    :param bool compressed: Indicates the use of compression.
    """

    security_suite: int = attr.ib(validator=[validate_security_suite_number])
    authenticated: bool = attr.ib(default=False)
    encrypted: bool = attr.ib(default=False)
    broadcast_key: bool = attr.ib(default=False)
    compressed: bool = attr.ib(default=False)

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        val = int.from_bytes(source_bytes, "big")  # just one byte.
        _security_suite = val & 0b00001111
        _authenticated = bool(val & 0b00010000)
        _encrypted = bool(val & 0b00100000)
        _key_set = bool(val & 0b01000000)
        _compressed = bool(val & 0b10000000)
        return cls(_security_suite, _authenticated, _encrypted, _key_set, _compressed)

    def to_bytes(self):
        _byte = self.security_suite
        if self.authenticated:
            _byte += 0b00010000
        if self.encrypted:
            _byte += 0b00100000
        if self.broadcast_key:
            _byte += 0b01000000
        if self.compressed:
            _byte += 0b10000000

        return _byte.to_bytes(1, "big")


def validate_key(suite: int, key: bytes) -> None:
    key_lengths = {0: 16, 1: 16, 2: 32}
    if len(key) != key_lengths[suite]:
        raise ValueError(
            f"Key with length {len(key)} is not the correct length for use with "
            f"security suite {suite}"
        )


# TODO: Is there a reason to support only encrypted or only authenthicated data?
#   only encrypted: additonal_data = b"". Dont add tag.
#   only authenticated: additional_data = security_control + auth_key + plain_text


def encrypt(
    security_control: SecurityControlField,
    system_title: bytes,
    invocation_counter: int,
    key: bytes,
    plain_text: bytes,
    auth_key: bytes,
) -> bytes:
    """
    Encrypts bytes according the to security context.
    """

    if not security_control.encrypted and not security_control.authenticated:
        raise NotImplementedError("encrypt() only handles authenticated encryption")

    if len(system_title) != 8:
        raise ValueError(f"System Title must be of lenght 8, not {len(system_title)}")

    # initialization vector is 12 bytes long and consists of the system_title (8 bytes)
    # and invocation_counter (4 bytes)
    iv = system_title + invocation_counter.to_bytes(4, "big")

    # Making sure the keys are of correct length for specified security suite
    validate_key(security_control.security_suite, key)
    validate_key(security_control.security_suite, auth_key)

    # Construct an AES-GCM Cipher object with the given key and iv. Allow for
    # truncating the auth tag
    encryptor = Cipher(
        algorithms.AES(key),
        modes.GCM(initialization_vector=iv, tag=None, min_tag_length=TAG_LENGTH),
    ).encryptor()

    # associated_data will be authenticated but not encrypted,
    # it must also be passed in on decryption.
    associated_data = security_control.to_bytes() + auth_key
    encryptor.authenticate_additional_data(associated_data)

    # Encrypt the plaintext and get the associated ciphertext.
    # GCM does not require padding.
    ciphertext = encryptor.update(plain_text) + encryptor.finalize()

    # dlms uses a tag lenght of 12 not the default of 16. Since we have set the minimum
    # tag length to 12 it is ok to truncated the tag down to 12 bytes.
    tag = encryptor.tag[:TAG_LENGTH]

    return ciphertext + tag


def decrypt(
    security_control: SecurityControlField,
    system_title: bytes,
    invocation_counter: int,
    key: bytes,
    cipher_text: bytes,
    auth_key: bytes,
):
    """
    Decrypts bytes according to the security context.
    """
    if not security_control.encrypted and not security_control.authenticated:
        raise NotImplementedError("encrypt() only handles authenticated encryption")

    if len(system_title) != 8:
        raise ValueError(f"System Title must be of lenght 8, not {len(system_title)}")

    # initialization vector is 12 bytes long and consists of the system_title (8 bytes)
    # and invocation_counter (4 bytes)
    iv = system_title + invocation_counter.to_bytes(4, "big")

    # Making sure the keys are of correct length for specified security suite
    validate_key(security_control.security_suite, key)
    validate_key(security_control.security_suite, auth_key)

    # extract the tag from the end of the cipher_text
    tag = cipher_text[-12:]
    ciphertext = cipher_text[:-12]
    try:
        # Construct a Cipher object, with the key, iv, and additionally the
        # GCM tag used for authenticating the message.
        decryptor = Cipher(
            algorithms.AES(key), modes.GCM(iv, tag, min_tag_length=12)
        ).decryptor()

        # We put associated_data back in or the tag will fail to verify
        # when we finalize the decryptor.
        associated_data = security_control.to_bytes() + auth_key
        decryptor.authenticate_additional_data(associated_data)

        # Decryption gets us the authenticated plaintext.
        # If the tag does not match an InvalidTag exception will be raised.
        return decryptor.update(ciphertext) + decryptor.finalize()
    except InvalidTag:
        raise DecryptionError(
            "Unable to decrypt ciphertext. Authentication tag is not valid. Ciphered "
            "text might have been tampered with or key, auth key, security control or "
            "invocation counter is wrong"
        )


def gmac(
    security_control: SecurityControlField,
    system_title: bytes,
    invocation_counter: int,
    key: bytes,
    auth_key: bytes,
    challenge: bytes,
):
    """
    GMAC is quite simply GCM mode where all data is supplied as additional
    authenticated data.
    If the GCM input is restricted to data that is not to be encrypted, the resulting
    specialization of GCM, called GMAC, is simply an authentication mode on the input
    data.
    """
    if security_control.encrypted:
        raise CipheringError(
            "Security for GMAC is set to encrypted, but this is not a "
            "valid choice since GMAC only authenticates  "
        )

    if len(system_title) != 8:
        raise ValueError(f"System Title must be of lenght 8, not {len(system_title)}")

    # initialization vector is 12 bytes long and consists of the system_title (8 bytes)
    # and invocation_counter (4 bytes)
    iv = system_title + invocation_counter.to_bytes(4, "big")

    # Making sure the keys are of correct length for specified security suite
    validate_key(security_control.security_suite, key)
    validate_key(security_control.security_suite, auth_key)

    # Construct an AES-GCM Cipher object with the given key and iv
    encryptor = Cipher(
        algorithms.AES(key),
        modes.GCM(initialization_vector=iv, tag=None, min_tag_length=TAG_LENGTH),
    ).encryptor()

    # associated_data will be authenticated but not encrypted,
    # so we put all data in the associated data.
    associated_data = security_control.to_bytes() + auth_key + challenge
    encryptor.authenticate_additional_data(associated_data)

    # Making sure to add an empty byte string as input. Then it will only be the
    # associated_data that will be authenticated.
    ciphertext = encryptor.update(b"") + encryptor.finalize()

    # We want the tag as it is the authenticated data. Need to truncated it first
    tag = encryptor.tag[:TAG_LENGTH]

    # ciphertext is really b"" here.
    return ciphertext + tag


def wrap_key(
    security_control: SecurityControlField, wrapping_key: bytes, key_to_wrap: bytes
):
    """
    Simple function to wrap a key for transfer
    """
    validate_key(security_control.security_suite, wrapping_key)
    validate_key(security_control.security_suite, key_to_wrap)
    wrapped_key = aes_key_wrap(wrapping_key, key_to_wrap)
    return wrapped_key


def unwrap_key(
    security_control: SecurityControlField, wrapping_key: bytes, wrapped_key: bytes
):
    """
    Simple function to unwrap a key received.
    """
    validate_key(security_control.security_suite, wrapping_key)
    validate_key(security_control.security_suite, wrapped_key)
    unwrapped_key = aes_key_unwrap(wrapping_key, wrapped_key)
    return unwrapped_key


def make_client_to_server_challenge(length: int = 8) -> bytes:
    """
    Return a valid challenge depending on the authentocation method.
    """
    if 8 <= length <= 64:
        return os.urandom(length)
    else:
        raise ValueError(
            f"Client to server challenge must be between 8 and 64 bytes. Got {length}"
        )


class AuthenticationMethodManager(Protocol):

    """

    HLS:
    After sending the HLS reply to the meter the meter sends back the result of the
    client challenge in the ActionResponse. To make sure the meter has dont the HLS
    auth correctly we must validate the data.
    The data looks different depending on the HLS type
    """

    secret: Optional[bytes]
    authentication_method: ClassVar[enumerations.AuthenticationMechanism]

    def get_calling_authentication_value(self) -> bytes:
        ...

    def hls_generate_reply_data(self, connection: DlmsConnection) -> bytes:
        ...

    def hls_meter_data_is_valid(self, data: bytes, connection: DlmsConnection) -> bool:
        ...


@attr.s(auto_attribs=True)
class NoSecurityAuthentication:
    secret = None
    authentication_method = enumerations.AuthenticationMechanism.NONE

    def get_calling_authentication_value(self) -> Optional[bytes]:
        return None

    def hls_generate_reply_data(self, connection: DlmsConnection) -> bytes:
        raise RuntimeError("Cannot call HLS methods when using NoAuthentication")

    def hls_meter_data_is_valid(self, data: bytes, connection: DlmsConnection) -> bool:
        raise RuntimeError("Cannot call HLS methods when using NoAuthentication")


@attr.s(auto_attribs=True)
class LowLevelSecurityAuthentication:
    secret: Optional[bytes]
    authentication_method = enumerations.AuthenticationMechanism.LLS

    def get_calling_authentication_value(self) -> Optional[bytes]:
        return self.secret

    def hls_generate_reply_data(self, connection: DlmsConnection) -> bytes:
        raise RuntimeError(
            "Cannot call HLS methods when using Low Level Authentication"
        )

    def hls_meter_data_is_valid(self, data: bytes, connection: DlmsConnection) -> bool:
        raise RuntimeError(
            "Cannot call HLS methods when using Low Level Authentication"
        )


@attr.s(auto_attribs=True)
class HighLevelSecurityGmacAuthentication:
    """
    HLS_GMAC:
            SC + IC + GMAC(SC + AK + Challenge)
    """

    secret = None
    authentication_method = enumerations.AuthenticationMechanism.HLS_GMAC
    challenge_length: int = attr.ib(default=32)
    calling_authentication_value: Optional[bytes] = attr.ib(
        init=False,
        default=attr.Factory(
            lambda self: make_client_to_server_challenge(self.challenge_length),
            takes_self=True,
        ),
    )

    def get_calling_authentication_value(self) -> bytes:
        return self.calling_authentication_value

    def hls_generate_reply_data(self, connection: DlmsConnection) -> bytes:
        """
        When the meter has enterted the HLS procedure the client firsts sends a reply
        to the server (meter) challenge. It is done with an ActionRequest to the
        current LN Association object in the meter. Method 2, Reply_to_HLS.

        Depending on the HLS type the data looks a bit different

        :param connection:
        :return:
        """
        if not connection.meter_to_client_challenge:
            raise exceptions.LocalDlmsProtocolError("Meter has not send challenge")
        if not connection.global_encryption_key:
            raise ProtectionError(
                "Unable to create GMAC. Missing global_encryption_key"
            )
        if not connection.global_authentication_key:
            raise ProtectionError(
                "Unable to create GMAC. Missing global_authentication_key"
            )
        only_auth_security_control = SecurityControlField(
            security_suite=connection.security_suite,
            authenticated=True,
            encrypted=False,
        )

        gmac_result = gmac(
            security_control=only_auth_security_control,
            system_title=connection.client_system_title,
            invocation_counter=connection.client_invocation_counter,
            key=connection.global_encryption_key,
            auth_key=connection.global_authentication_key,
            challenge=connection.meter_to_client_challenge,
        )
        return (
            only_auth_security_control.to_bytes()
            + connection.client_invocation_counter.to_bytes(4, "big")
            + gmac_result
        )

    def hls_meter_data_is_valid(self, data: bytes, connection: DlmsConnection) -> bool:
        security_control = SecurityControlField.from_bytes(data[0].to_bytes(1, "big"))
        invocation_counter = int.from_bytes(data[1:5], "big")
        gmac_result = data[-12:]

        if not connection.global_encryption_key:
            raise ProtectionError(
                "Unable to verify GMAC. Missing global_encryption_key"
            )
        if not connection.global_authentication_key:
            raise ProtectionError(
                "Unable to verify GMAC. Missing global_authentication_key"
            )
        if not connection.meter_system_title:
            raise ProtectionError(
                "Unable to verify GMAC. Have not received the meters system title."
            )

        correct_gmac = gmac(
            security_control=security_control,
            system_title=connection.meter_system_title,
            invocation_counter=invocation_counter,
            key=connection.global_encryption_key,
            auth_key=connection.global_authentication_key,
            challenge=self.get_calling_authentication_value(),
        )
        return gmac_result == correct_gmac


@attr.s(auto_attribs=True)
class HighLevelSecurityCommonAuthentication:
    """
    In older meters that only specify auth method 2 it is common to use AES128-ECB to
    encrypt the client and meter challanges in the reply-to-HLS flow

    """

    secret: bytes
    authentication_method = enumerations.AuthenticationMechanism.HLS

    @property
    def padded_secret(self) -> bytes:
        """
        To be able to use AES128 our encryption key must be 128 bits 16 bytes long
        """
        to_pad = 16 - len(self.secret)
        padding = bytes(to_pad)
        return self.secret + padding

    def get_calling_authentication_value(self) -> bytes:
        return self.secret

    def hls_generate_reply_data(self, connection: DlmsConnection) -> bytes:
        encryptor = Cipher(algorithms.AES(self.padded_secret), modes.ECB()).encryptor()
        return (
            encryptor.update(connection.meter_to_client_challenge)
            + encryptor.finalize()
        )

    def hls_meter_data_is_valid(self, data: bytes, connection: DlmsConnection) -> bool:
        encryptor = Cipher(algorithms.AES(self.padded_secret), modes.ECB()).encryptor()
        calculated_data = (
            encryptor.update(self.get_calling_authentication_value())
            + encryptor.finalize()
        )
        return data == calculated_data



from typing import *

import attr

from dlms_cosem import enumerations
from dlms_cosem.cosem.obis import Obis


@attr.s(auto_attribs=True)
class CosemAttribute:

    interface: enumerations.CosemInterface
    instance: Obis
    attribute: int

    LENGTH: ClassVar[int] = 2 + 6 + 1

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        if len(source_bytes) != cls.LENGTH:
            raise ValueError(
                f"Data is not of correct lenght. Should be {cls.LENGTH} but is "
                f"{len(source_bytes)}"
            )
        interface = enumerations.CosemInterface(int.from_bytes(source_bytes[:2], "big"))
        instance = Obis.from_bytes(source_bytes[2:8])
        attribute = source_bytes[-1]
        return cls(interface, instance, attribute)

    def to_bytes(self) -> bytes:
        return b"".join(
            [
                self.interface.to_bytes(2, "big"),
                self.instance.to_bytes(),
                self.attribute.to_bytes(1, "big"),
            ]
        )


@attr.s(auto_attribs=True)
class CosemMethod:

    interface: enumerations.CosemInterface
    instance: Obis
    method: int

    LENGTH: ClassVar[int] = 2 + 6 + 1

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        if len(source_bytes) != cls.LENGTH:
            raise ValueError(
                f"Data is not of correct length. Should be {cls.LENGTH} but is "
                f"{len(source_bytes)}"
            )
        interface = enumerations.CosemInterface(int.from_bytes(source_bytes[:2], "big"))
        instance = Obis.from_bytes(source_bytes[2:8])
        method = source_bytes[-1]
        return cls(interface, instance, method)

    def to_bytes(self) -> bytes:
        return b"".join(
            [
                self.interface.to_bytes(2, "big"),
                self.instance.to_bytes(),
                self.method.to_bytes(1, "big"),
            ]
        )



import attr
import structlog

from dlms_cosem.hdlc import frames
from dlms_cosem.hdlc.exceptions import LocalProtocolError

LOG = structlog.get_logger()


class _SentinelBase(type):
    """
    Sentinel values

     - Inherit identity-based comparison and hashing from object
     - Have a nice repr
     - Have a *bonus property*: type(sentinel) is sentinel

     The bonus property is useful if you want to take the return value from
     next_event() and do some sort of dispatch based on type(event).

     Taken from h11.
    """

    def __repr__(self):
        return self.__name__


def make_sentinel(name):
    cls = _SentinelBase(name, (_SentinelBase,), {})
    cls.__class__ = cls
    return cls


# NOT_CONNECTED is when we have created a session but not actually set up HDLC
# connection with the server (meter). We used a SNMR frame to set up the connection
NOT_CONNECTED = make_sentinel("NOT_CONNECTED")

# IDLE State is when we are connected but we have not started a data exchange or we
# just finished a data exchange
IDLE = make_sentinel("IDLE")

AWAITING_RESPONSE = make_sentinel("AWAITING_RESPONSE")


AWAITING_CONNECTION = make_sentinel("AWAITING_CONNECTION")

AWAITING_DISCONNECT = make_sentinel("AWAITING_DISCONNECT")

CLOSED = make_sentinel("CLOSED")

NEED_DATA = make_sentinel("NEED_DATA")

# TODO: segmentation handling is not working with this state layout.

HDLC_STATE_TRANSITIONS = {
    NOT_CONNECTED: {frames.SetNormalResponseModeFrame: AWAITING_CONNECTION},
    AWAITING_CONNECTION: {frames.UnNumberedAcknowledgmentFrame: IDLE},
    IDLE: {
        frames.InformationFrame: AWAITING_RESPONSE,
        frames.DisconnectFrame: AWAITING_DISCONNECT,
        frames.ReceiveReadyFrame: AWAITING_RESPONSE,
    },
    AWAITING_RESPONSE: {frames.InformationFrame: IDLE, frames.ReceiveReadyFrame: IDLE},
    AWAITING_DISCONNECT: {frames.UnNumberedAcknowledgmentFrame: NOT_CONNECTED},
}


SEND_STATES = [NOT_CONNECTED, IDLE]
RECEIVE_STATES = [AWAITING_CONNECTION, AWAITING_RESPONSE, AWAITING_DISCONNECT]

# TODO: does the ssn and rsn belong in the state? Comparing to H11 that is only
#   using types in the state not full objects. Maybe it should be stored on the
#   connection?


@attr.s(auto_attribs=True)
class HdlcConnectionState:
    """
    Handles state changes in HDLC, we only focus on Client implementation as of now.

    A HDLC frame is passed to `process_frame` and it moves the state machine to the
    correct state. If a frame is processed that is not set to be able to transition
    the state in the current state a LocalProtocolError is raised.
    """

    current_state: _SentinelBase = attr.ib(default=NOT_CONNECTED)

    def process_frame(self, frame):

        self._transition_state(type(frame))

    def _transition_state(self, frame_type):
        try:
            new_state = HDLC_STATE_TRANSITIONS[self.current_state][frame_type]
        except KeyError:
            raise LocalProtocolError(
                f"can't handle frame type {frame_type} when state={self.current_state}"
            )
        old_state = self.current_state
        self.current_state = new_state
        LOG.debug(f"HDLC state transitioned", old_state=old_state, new_state=new_state)



from enum import IntEnum
from typing import *

import attr

from dlms_cosem import enumerations


@attr.s(auto_attribs=True)
class ExceptionResponse:
    TAG: ClassVar[int] = 216

    state_error: enumerations.StateException
    service_error: enumerations.ServiceException
    invocation_counter_data: Optional[int] = attr.ib(default=None)

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(
                f"Tag for ExceptionResponse is not {cls.TAG}. Got {tag} instead."
            )
        state_error = enumerations.StateException(data.pop(0))
        service_error = enumerations.ServiceException(data.pop(0))

        if service_error == enumerations.ServiceException.INVOCATION_COUNTER_ERROR:
            invocation_counter_data = int.from_bytes(data, "big")
        else:
            invocation_counter_data = None

        return cls(state_error, service_error, invocation_counter_data)

    def to_bytes(self):
        if not self.invocation_counter_data:
            return bytes([self.TAG, self.state_error, self.service_error])
        return bytes(
            [
                self.TAG,
                self.state_error,
                self.service_error,
                self.invocation_counter_data,
            ]
        )



from enum import IntEnum, unique


@unique
class DataAccessResult(IntEnum):
    SUCCESS = 0
    HARDWARE_FAULT = 1
    TEMPORARY_FAILURE = 2
    READ_WRITE_DENIED = 3
    OBJECT_UNDEFINED = 4
    OBJECT_CLASS_INCONSISTENT = 9
    OBJECT_UNAVAILABLE = 11
    TYPE_UNMATCHED = 12
    SCOPE_OF_ACCESS_VIOLATED = 13
    DATA_BLOCK_UNAVAILABLE = 14
    LONG_GET_ABORTED = 15
    NO_LONG_GET_IN_PROGRESS = 16
    LONG_SET_ABORTED = 17
    NO_LONG_SET_IN_PROGRESS = 18
    DATA_BLOCK_NUMBER_INVALID = 19
    OTHER_REASON = 250


class GetRequestType(IntEnum):
    NORMAL = 1
    NEXT = 2
    WITH_LIST = 3


class GetResponseType(IntEnum):
    NORMAL = 1
    WITH_BLOCK = 2
    WITH_LIST = 3


class SetRequestType(IntEnum):
    NORMAL = 1
    WITH_FIRST_BLOCK = 2
    WITH_BLOCK = 3
    WITH_LIST = 4
    FIRST_BLOCK_WITH_LIST = 5


class SetResponseType(IntEnum):
    NORMAL = 1
    WITH_BLOCK = 2
    WITH_LAST_BLOCK = 3
    LAST_BLOCK_WITH_LIST = 4
    WITH_LIST = 5


class ActionType(IntEnum):
    NORMAL = 1
    NEXT_PBLOCK = 2
    WITH_LIST = 3
    WITH_FIRST_PBLOCK = 4
    WITH_LIST_AND_FIRST_PBLOCK = 5
    WITH_PBLOCK = 6


class StateException(IntEnum):
    SERVICE_NOT_ALLOWED = 1
    SERVICE_UNKNOWN = 2


class ServiceException(IntEnum):
    OPERATION_NOT_POSSIBLE = 1
    SERVICE_NOT_SUPPORTED = 2
    OTHER_REASON = 3
    PDU_TOO_LONG = 4
    DECIPHERING_ERROR = 5
    INVOCATION_COUNTER_ERROR = 6


class ApplicationReferenceError(IntEnum):
    OTHER = 0
    TIME_ELAPSED = 1  # timeout since request sent
    APPLICATION_UNREACHABLE = 2  # peer AEi not reachable
    APPLICATION_REFERENCE_INVALID = 3  # addressing problems
    APPLICATION_CONTEXT_UNSUPPORTED = 4  # incompatible application context
    PROVIDER_COMMUNICATION_ERROR = 5  # error in local or remote equipment
    DECIPHERING_ERROR = 6  # Error detected in deciphering function.


class HardwareResourceError(IntEnum):
    OTHER = 0
    MEMORY_UNAVAILABLE = 1
    PROCESSOR_RESOURCE_UNAVAILABLE = 2
    MASS_STORAGE_UNAVAILABLE = 3
    OTHER_RESOURCE_UNAVAILABLE = 4


class VdeStateError(IntEnum):
    OTHER = 0
    NO_DLMS_CONTEXT = 1
    LOADING_DATASET = 2
    STATUS_NO_CHANGE = 3
    STATUS_INOPERABLE = 4


class ServiceError(IntEnum):
    OTHER = 0
    PDU_SIZE = 1  # PDU too long
    SERVICE_UNSUPPORTED = 2  # Service unsupported as in conformance block


class DefinitionError(IntEnum):
    OTHER = 0
    OBJECT_UNDEFINED = 1  # object not defined at the VDE
    OBJECT_CLASS_INCONSISTENT = 2  # class of object incompatible with asked service
    OBJECT_ATTRIBUTE_INCONSISTENT = 3  # object attributes are inconsistent to doc


class AccessError(IntEnum):
    OTHER = 0
    SCOPE_OF_ACCESS_VIOLATED = 1  # access denied through authorisation reason
    OBJECT_ACCESS_VIOLATED = 2  # access incompatible with object attribute
    HARDWARE_FAULT = 3  # access fail for hardware reasons
    OBJECT_UNAVAILABLE = 4  # VDE hands object for unavailable


class InitiateError(IntEnum):
    OTHER = 0
    DLMS_VERSION_TOO_LOW = 1  # proposed dlms version is too low
    INCOMPATIBLE_CONFORMANCE = 2  # proposed service not sufficient
    PDU_SIZE_TOO_SHORT = 3  # proposed pdu size is too short
    REFUSED_BY_VDE_HANDLER = 4  # vaa creation impossible or not allowed


class LoadDataError(IntEnum):
    OTHER = 0
    PRIMITIVE_OUT_OF_SEQUENCE = 1
    NOT_LOADABLE = 2
    DATASET_SIZE_TOO_LARGE = 3
    NOT_AWAITED_SEGMENT = 4
    INTERPRETATION_FAILURE = 5
    STORAGE_FAILURE = 6
    DATASET_NOT_READY = 7


class DataScopeError(IntEnum):
    OTHER = 0


class TaskError(IntEnum):
    OTHER = 0
    NO_REMOTE_CONTROL = 1
    TI_STOPPED = 2
    TI_RUNNING = 3
    TI_UNUSABLE = 4


class OtherError(IntEnum):
    OTHER = 0


@unique
class CosemInterface(IntEnum):

    # Parameters and measurement data.
    DATA = 1
    REGISTER = 3
    EXTENDED_REGISTER = 4
    DEMAND_REGISTER = 5
    REGISTER_ACTIVATION = 6
    PROFILE_GENERIC = 7
    UTILITY_TABLES = 26
    REGISTER_TABLE = 61
    COMPACT_DATA = 62
    STATUS_MAPPING = 63

    # Access control and management
    ASSOCIATION_SN = 12
    ASSOCIATION_LN = 15
    SAP_ASSIGNMENT = 17
    IMAGE_TRANSFER = 18
    SECURITY_SETUP = 64
    PUSH = 40
    COSEM_DATA_PROTECTION = 30
    FUNCTION_CONTROL = 122
    ARRAY_MANAGER = 123
    COMMUNICATION_PORT_PROTECTION = 124

    # Time and event bound control
    CLOCK = 8
    SCRIPT_TABLE = 9
    SCHEDULE = 10
    SPECIAL_DAYS_TABLE = 11
    ACTIVITY_CALENDAR = 20
    REGISTER_MONITOR = 21
    SINGLE_ACTION_SCHEDULE = 22
    DISCONNECT_CONTROL = 70
    LIMITER = 71
    PARAMETER_MONITOR = 65
    SENSOR_MANAGER = 67
    ARBITRATOR = 68

    # Payment related interfaces
    ACCOUNT = 111
    CREDIT = 112
    CHARGE = 113
    TOKEN_GATEWAY = 115

    # Data exchange over local ports and modems
    IEC_LOCAL_PORT_SETUP = 19
    IEC_HDLC_SETUP = 23
    IEC_TWISTED_PAIR_SETUP = 24
    MODEM_CONFIGURATION = 27
    AUTO_ANSWER = 28
    AUTO_CONNECT = 29
    GPRS_MODEM_SETUP = 45
    GSM_DIAGNOSTICS = 47
    LTE_MONITORING = 151

    # Data exchange over M-Bus
    MBUS_SLAVE_PORT_SETUP = 25
    MBUS_CLIENT = 72
    MBUS_WIRELESS_MODE_Q_CHANNEL = 73
    MBUS_MASTER_PORT_SETUP = 74
    MBUS_PORT_SETUP_DLMS_COSEM_SERVER = 76
    MBUS_DIAGNOSTICS = 77

    # Data exchange over Internet
    TCP_UDP_SETUP = 41
    IPV4_SETUP = 42
    IPV6_SETUP = 48
    MAC_ADDRESS_SETUP = 43
    PPP_SETUP = 44
    SMTP_SETUP = 46
    NTP_SETUP = 100

    # Data exchange using S-FSK PLC
    S_FSK_PHY_MAC_SETUP = 50
    S_FSK_ACTIVE_INITIATOR = 51
    S_FSK_MAC_SYNCHRONISATION_TIMEOUTS = 52
    S_FSK_MAC_COUNTERS = 53
    S_FSK_IEC_61334_4_32_LLC_SETUP = 55
    S_FSK_REPORTING_SYSTEM_LIST = 56

    # LLC layers for IEC 8802-2
    IEC_8802_2_LLC_TYPE_1_SETUP = 57
    IEC_8802_2_LLC_TYPE_2_SETUP = 58
    IEC_8802_2_LLC_TYPE_3_SETUP = 59

    # Narrowband OFDM PLC profile for PRIME networks
    PRIME_61344_4_32_LLC_SSCS_SETUP = 80
    PRIME_OFDM_PLC_PHYSICAL_LAYER_COUNTERS = 81
    PRIME_OFDM_PLC_MAC_SETUP = 82
    PRIME_OFDM_PLC_MAC_FUNCTIONAL_PARAMETERS = 83
    PRIME_OFDM_PLC_MAC_COUNTERS = 84
    PRIME_OFDM_PLC_MAC_NETWORK_ADMINISTRATION_DATA = 85
    # PRIME_OFDM_PLC_MAC_ADDRESS_SETUP = 43
    PRIME_OFDM_PLC_MAC_APPLICATION_IDENTIFICATION = 86

    # Narrowband OFDM PLC profile for G3-PLC network
    G3_PLC_MAC_LAYER_COUNTERS = 90
    G3_PLC_MAC_SETUP = 91
    G3_PLC_6LOWPAN_ADAPTATION_LAYER_SETUP = 92

    # HS-PLC IEC 12139-1
    HS_PLC_IEC_12139_1_MAC_SETUP = 140
    HS_PLC_IEC_12139_1_CPAS_SETUP = 141
    HS_PLC_IEC_12139_1_IP_SSAS_SETUP = 142
    HS_PLC_IEC_12139_1_HDLC_SSAS_SETUP = 143

    # Zigbee
    ZIGBEE_SAS_STARTUP = 101
    ZIGBEE_SAS_JOIN = 102
    ZIGBEE_SAS_APS_FRAGMENTATION = 103
    ZIGBEE_NETWORK_CONTROL = 104
    ZIGBEE_TUNNEL_SETUP = 105

    # LPWAN networks
    SCHC_LPWAN = 126
    SCHC_LPWAN_DIAGNOSTICS = 127
    LORAWAN_SETUP = 128
    LORAWAN_DIAGNOSTICS = 129

    # Wi-SUN
    WISUN_SETUP = 95
    WISUM_DIAGNOSTICS = 96
    RPL_DIAGNOSTICS = 97
    MPL_DIAGNOSTICS = 98

    # IEC 14908 PLC
    IEC_14908_IDENTIFICATION = 130
    IEC_14908_PROTOCOL_SETUP = 131
    IEC_14908_PROTOCOL_STATUS = 132
    IEC_14908_DIAGNOSTICS = 133

    # TODO: how do we represent different versions of interface classes.
    #   I guess the python class representing the interface should have the versions
    #   Like ProfileGenericV1, ProfileGerericV2


@unique
class ReleaseRequestReason(IntEnum):
    NORMAL = 0
    URGENT = 1
    USER_DEFINED = 30


@unique
class ReleaseResponseReason(IntEnum):
    NORMAL = 0
    NOT_FINISHED = 1
    USER_DEFINED = 30


class AuthenticationMechanism(IntEnum):
    NONE = 0
    LLS = 1
    HLS = 2
    HLS_MD5 = 3  # Insecure. Don't use with new meters
    HLS_SHA1 = 4  # Insecure. Don't use with new meters
    HLS_GMAC = 5
    HLS_SHA256 = 6
    HLS_ECDSA = 7


class AcseServiceUserDiagnostics(IntEnum):
    NULL = 0
    NO_REASON_GIVEN = 1
    APPLICATION_CONTEXT_NAME_NOT_SUPPORTED = 2
    CALLING_AP_TITLE_NOT_RECOGNIZED = 3
    CALLING_AP_INVOCATION_IDENTIFIER_NOT_RECOGNIZED = 4
    CALLING_AE_QUALIFIER_NOT_RECOGNIZED = 5
    CALLING_AE_INVOCATION_IDENTIFIER_NOT_RECOGNIZED = 6
    CALLED_AP_TITLE_NOT_RECOGNIZED = 7
    CALLED_AP_INVOCATION_IDENTIFIER_NOT_RECOGNIZED = 8
    CALLED_AE_QUALIFIER_NOT_RECOGNIZED = 9
    CALLED_AE_INVOCATION_IDENTIFIER_NOT_RECOGNIZED = 10
    AUTHENTICATION_MECHANISM_NAME_NOT_RECOGNIZED = 11
    AUTHENTICATION_MECHANISM_NAME_REQUIRED = 12
    AUTHENTICATION_FAILED = 13
    AUTHENTICATION_REQUIRED = 14


class AcseServiceProviderDiagnostics(IntEnum):
    NULL = 0
    NO_REASON_GIVEN = 1
    NO_COMMON_ACSE_VERSION = 2


class AssociationResult(IntEnum):
    ACCEPTED = 0
    REJECTED_PERMANENT = 1
    REJECTED_TRANSIENT = 2
    # TODO: What does transient rejection mean?


class ActionResultStatus(IntEnum):
    SUCCESS = 0
    HARDWARE_FAULT = 1
    TEMPORARY_FAILURE = 2
    READ_WRITE_DENIED = 3
    OBJECT_UNDEFINED = 4
    OBJECT_CLASS_INCONSISTENT = 9
    OBJECT_UNAVAILABLE = 11
    TYPE_UNMATCHED = 12
    SCOPE_OF_ACCESS_VIOLATED = 13
    DATA_BLOCK_UNAVAILABLE = 14
    LONG_ACTION_ABORTED = 15
    NO_LONG_ACTION_IN_PROGRESS = 16
    OTHER_REASON = 250



from datetime import date, datetime, time, timedelta, timezone
from typing import *

import attr
from dateutil.tz import tzoffset

"""
About datetimes in DLMS.

Date,time and date-time may be represented simply as Octetstring (Tag=9) where the length
shows what type of data it is.
* 5 bytes = date
* 4 bytes = time
* 12 bytes = datetime.

Date, time and datetime may also be represented by concise data classes of Date
(tag=26), Time (tag=27) and DateTime (tag=25).

The special cases seems to mostly be used to define "cron-like" date. For example
if you want to define a date for every year you would leave year undefined.
Initially we will ignore this functionality.

"""


@attr.s(auto_attribs=True)
class ClockStatus:
    """
    :parameter invalid: Time could not be recovered after incident. Manufacturer dependant
        Shall not be set if `doubtful` is set.
    :parameter doubtful: Time could be recovered after incidnet, but correctness is not
        guaranteed. Manufacturer specific.
        Shall not be set if `invalid` is set.
    :parameter different_base: Set if timing source is different from the one
        specified in clock_base. (secondary clock for example)
    :parameter invalid_status: Something in the status itself is invalid.
        Manufacturer specific.
    :parameter daylight_saving_active: indicates if datetime contings daylight savings
        deviation (summer time)

    """

    invalid: bool = attr.ib(default=False)
    doubtful: bool = attr.ib(default=False)
    different_base: bool = attr.ib(default=False)
    invalid_status: bool = attr.ib(default=False)
    daylight_saving_active: bool = attr.ib(default=False)

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        if len(source_bytes) != 1:
            raise ValueError(f"ClockStatus is of 1 bytes, got: {len(source_bytes)}")
        value = int.from_bytes(source_bytes, "big")
        invalid = bool(value & 0b00000001)
        doubtful = bool(value & 0b00000010)
        different_base = bool(value & 0b00000100)
        invalid_status = bool(value & 0b00001000)
        daylight_saving_active = bool(value & 0b10000000)

        return cls(
            invalid, doubtful, different_base, invalid_status, daylight_saving_active
        )

    def to_bytes(self):
        value = 0
        if self.invalid:
            value += 0b00000001
        if self.doubtful:
            value += 0b00000010
        if self.different_base:
            value += 0b00000100
        if self.invalid_status:
            value += 0b00001000
        if self.daylight_saving_active:
            value += 0b10000000

        return value.to_bytes(1, "big")


def validate_day(value: Optional[int]):
    if value:
        if 1 > value > 31:
            raise ValueError(f"Day can only be within 1-31")


def validate_month(value: Optional[int]):
    if value:
        if 1 > value > 12:
            raise ValueError(f"Month can only be within 1-12")


def validate_weekday(value: Optional[int]):
    if value:
        if 1 > value > 7:
            raise ValueError(f"Day can only be within 1-7")


def validate_hour(value: Optional[int]):
    if value:
        if 0 > value > 23:
            raise ValueError(f"Minutes and seconds can only be within 0-23")


def validate_minute_or_second(value: Optional[int]):
    if value:
        if 0 > value > 59:
            raise ValueError(f"Minutes and seconds can only be within 0-59")


def validate_hundredths(value: Optional[int]):
    if value:
        if 0 > value > 99:
            raise ValueError(f"Hundredths can only be within 0-59")


def get_optional_value(
    value: Union[bytes, int],
    optional_indicator: bytes,
    replace_with: Optional[int] = None,
    signed: bool = False,
) -> Optional[int]:
    if isinstance(value, bytes):
        if value == optional_indicator:
            return replace_with
    else:
        if value == int.from_bytes(optional_indicator, "big", signed=signed):
            return replace_with
    return value


def date_from_bytes(source_bytes: bytes) -> date:

    """
    Date is represented by 5 bytes.
    [year highbyte, year lowbyte, month, day of month, day of week

    year: long-unsigned int.
        Special case:
            0xFFFF == not specified.

    month:  unsigned int (1-12) 1 = January.
        Special cases:
            0xfd = daylight_savings_end
            0xfe = daylight_savings_begin
            0xff = not specified.

    day_of_month: unsigned int 1-31,
        Special cases:
            0xe0-0xfc: reserved
            0xfd = 2nd last day of month
            0xfe = last day of month
            0xff = not specified

    day_of_week: unsigned in 1-7, 1 = Monday
        Special cases:
            0xff = not specifed

    The elements dayOfMonth and dayOfWeek shall be interpreted together:
    - if last dayOfMonth is specified (0xFE) and dayOfWeek is wildcard, this specifies
        the last calendar day of the month;
    - if last dayOfMonth is specified (0xFE) and an explicit dayOfWeek is specified
        (for example 7, Sunday) then it is the last occurrence of the weekday specified
        in the month, i.e. the last Sunday;
    - if the year is not specified (0xFFFF), and dayOfMonth and dayOfWeek are both
        explicitly specified, this shall be interpreted as the dayOfWeek on, or
        following dayOfMonth;
    - if the year and month are specified, and both the dayOfMonth and dayOfWeek are
        explicitly specified but the values are not consistent it shall be considered
        as an error.



    """
    if len(source_bytes) != 5:
        raise ValueError(f"Date is represented by 5 bytes, but got {len(source_bytes)}")

    year = get_optional_value(int.from_bytes(source_bytes[:2], "big"), b"\xff\xff")
    month = get_optional_value(source_bytes[2], b"\xff")
    day_of_month = get_optional_value(source_bytes[3], b"\xff")
    day_of_week = get_optional_value(source_bytes[4], b"\xff")
    validate_month(month)
    validate_day(day_of_month)
    validate_weekday(day_of_week)

    return date(year=year, month=month, day=day_of_month)


def time_from_bytes(source_bytes: bytes) -> time:
    """
      Time is represented by 4 bytes.
    [hour, minute, second, hundredths]

    hour: unsigned int (0-23)
        Special case:
            0xFF == not specified.

    minute:  unsigned int (0-59)
        Special cases:
            0xff = not specified.

    second: unsigned int (0-59),
        Special cases:
            0xff = not specified

    hundredths: unsigned in (0-99)
        Special cases:
            0xff = not specifed

    """

    if len(source_bytes) != 4:
        raise ValueError(f"Time is represented by 4 bytes, but got {len(source_bytes)}")

    hour: int = get_optional_value(source_bytes[0], b"\xff", replace_with=0)
    minute: int = get_optional_value(source_bytes[1], b"\xff", replace_with=0)
    seconds: int = get_optional_value(source_bytes[2], b"\xff", replace_with=0)
    hundredths: int = get_optional_value(source_bytes[3], b"\xff", replace_with=0)
    validate_hour(hour)
    validate_minute_or_second(minute)
    validate_minute_or_second(seconds)
    validate_hundredths(hundredths)

    return time(
        hour=hour, minute=minute, second=seconds, microsecond=hundredths * 10000
    )


def utc_offset_minutes(offset_minutes: Optional[int]) -> Optional[tzoffset]:
    """
    Big issue in DLMS about timezone.
    The DLMS standard and IDIS standard use the "correct" way of defining the utc offset.
    In the Blue Book 4.1.6.1 the timezone deviation is defined as minutes from local time to UTC.
    NOT deviation from UTC.
    In practice that means we need to negate the offset.
    UTC+01:00 == -60 minutes since you need to subtract 60 minutes to get to UTC.
    UTC-01:00 == +60 minutes since you need to add 60 minutes to get to UTC.

    To make it harder some companion standard is not using the the standard way of
    deviation from localtime but the deviation from UTC.

    # TODO: We need a way to handle different ways of interpretating the timezone offset.

    """
    if offset_minutes:
        return tzoffset(name=None, offset=-(offset_minutes * 60))
    else:
        return None


def datetime_from_bytes(source_bytes: bytes) -> Tuple[datetime, Optional[ClockStatus]]:
    """
     Datetime is represented byte 12 bytes
     [date[year highbyte, year lowbyte, month, day of month, day of week],
     time[hour, minute, second, hundredths], deviation_high, deviation_low, clock_status]
     }

    date: as above
    time: as above
    deviation: deviation from UTC in minutes. Shows the timezone. signed long-integer
         (-720 < dev < 720)
         Special case:
             0x8000: not specified.

    """
    if len(source_bytes) != 12:
        raise ValueError(
            f"Datetime is represented by 12 bytes, but got {len(source_bytes)}"
        )
    d = date_from_bytes(source_bytes[:5])
    t = time_from_bytes(source_bytes[5:9])
    deviation = get_optional_value(
        int.from_bytes(source_bytes[9:11], "big", signed=True), b"\x80\x00", signed=True
    )
    status_bytes = source_bytes[-1].to_bytes(1, "big")
    status = ClockStatus.from_bytes(status_bytes) if status_bytes else None

    dt = datetime(
        year=d.year,
        month=d.month,
        day=d.day,
        hour=t.hour,
        minute=t.minute,
        second=t.second,
        microsecond=t.microsecond,
        tzinfo=utc_offset_minutes(deviation),
    )

    return dt, status


def date_to_bytes(d: date) -> bytes:
    """Will set day of week to unspecified. """

    year = d.year
    month = d.month
    day = d.day
    year_bytes = year.to_bytes(2, "big")
    month_byte = month.to_bytes(1, "big")
    day_byte = day.to_bytes(1, "big")
    day_of_week_unspecified = b"\xff"

    return year_bytes + month_byte + day_byte + day_of_week_unspecified


def time_to_bytes(t: time) -> bytes:

    return (
        t.hour.to_bytes(1, "big")
        + t.minute.to_bytes(1, "big")
        + t.second.to_bytes(1, "big")
        + int(t.microsecond / 10000).to_bytes(1, "big")
    )


def datetime_to_bytes(dt: datetime, clock_status: Optional[ClockStatus] = None):

    date_bytes = date_to_bytes(dt.date())
    time_bytes = time_to_bytes(dt.time())
    if dt.tzinfo is None:
        timezone_bytes = b"\x80\x00"
    else:
        # negating the offset to match dlms standard offset representation
        timezone_bytes = int(-(dt.utcoffset().total_seconds() / 60)).to_bytes(
            2, "big", signed=True
        )

    if clock_status is None:
        clock_status_bytes = ClockStatus().to_bytes()
    else:
        clock_status_bytes = clock_status.to_bytes()

    return date_bytes + time_bytes + timezone_bytes + clock_status_bytes



from ctypes import c_ushort
from typing import *


class CRCCCITT:
    """
    CRC CCITT - HDLC Style 16-bit
    In accordning with ANSI C12.18(2006)
    Using 0xFFFF as initial value
    Running over serial so all message bytes need to be reversed
    before calculation (because least significant bit is sent first)
    resulting crc bytes needs to be reversed to become in correct order
    The reversed crc is then XOR:ed with 0xFFFF

    """

    crc_ccitt_table: List[int] = list()

    # The CRC's are computed using polynomials.

    crc_ccitt_constant = 0x1021

    def __init__(self):
        self.starting_value = 0xFFFF

        # initialize the pre-calculated tables
        if not len(self.crc_ccitt_table):
            self.init_crc_table()

    def calculate_for(self, input_data, lsb_first=False) -> bytes:
        """

        :param input_data:
        :param lsb_first: Indicate if the Least significant byte should be returned
            first (little endian)
        :return:
        """

        # need to revers bits in bytes
        reversed_data = reverse_byte_message(input_data)

        reversed_crc = self._calculate(reversed_data)
        lsb_rev = reversed_crc & 0x00FF
        lsb: int = ord(reverse_byte(lsb_rev))
        lsb ^= 0xFF
        lsb_byte = lsb.to_bytes(1, "big")
        msb_rev = (reversed_crc & 0xFF00) >> 8
        msb = ord(reverse_byte(msb_rev))
        msb ^= 0xFF
        msb_byte = msb.to_bytes(1, "big")

        if lsb_first:
            return b"".join([lsb_byte, msb_byte])
        else:
            return b"".join([msb_byte, lsb_byte])

    def _calculate(self, input_data: bytes):

        crc_value = self.starting_value

        for c in input_data:
            tmp = ((crc_value >> 8) & 0xFF) ^ c
            crc_shifted = (crc_value << 8) & 0xFF00
            crc_value = crc_shifted ^ self.crc_ccitt_table[tmp]

        return crc_value

    def init_crc_table(self):
        """The algorithm uses tables with pre-calculated values"""
        for i in range(0, 256):
            crc = 0
            c = i << 8

            for j in range(0, 8):
                if (crc ^ c) & 0x8000:
                    crc = c_ushort(crc << 1).value ^ self.crc_ccitt_constant
                else:
                    crc = c_ushort(crc << 1).value

                c = c_ushort(c << 1).value  # equivalent of c = c << 1

            self.crc_ccitt_table.append(crc)


def reverse_byte(byte_to_reverse):
    and_value = 1
    reversed_byte = 0
    for i in range(0, 8):
        reversed_byte += ((byte_to_reverse & and_value) >> i) * (2 ** (7 - i))
        and_value += and_value

    return (chr(reversed_byte)).encode("latin-1")


def reverse_byte_message(msg):
    reversed_mgs = b""
    for char in msg:
        reversed_mgs += reverse_byte(char)
    return reversed_mgs



import pytest

from dlms_cosem import enumerations, state
from dlms_cosem.exceptions import LocalDlmsProtocolError
from dlms_cosem.protocol import acse
from dlms_cosem.protocol.acse import UserInformation
from dlms_cosem.protocol.xdlms import Conformance, InitiateRequest


def test_non_aarq_on_initial_raises_protocol_error():
    s = state.DlmsConnectionState()

    with pytest.raises(LocalDlmsProtocolError):
        s.process_event(acse.ReleaseResponse())


def test_aarq_makes_dlms_waiting_for_aare():
    s = state.DlmsConnectionState()
    s.process_event(
        acse.ApplicationAssociationRequest(
            user_information=UserInformation(
                InitiateRequest(proposed_conformance=Conformance())
            )
        )
    )
    assert s.current_state == state.AWAITING_ASSOCIATION_RESPONSE


def test_aare_sets_ready_on_waiting_aare_response():
    s = state.DlmsConnectionState(current_state=state.AWAITING_ASSOCIATION_RESPONSE)
    s.process_event(
        acse.ApplicationAssociationResponse(
            enumerations.AssociationResult.ACCEPTED,
            result_source_diagnostics=enumerations.AcseServiceUserDiagnostics.NULL,
        )
    )
    assert s.current_state == state.READY



import attr

from dlms_cosem import dlms_data, utils

from .base import CosemAttribute


@attr.s(auto_attribs=True)
class CaptureObject:
    """
    Definition of a value that is supposed to be saved in a Profile Generic.

    A data_index of 0 means the whole attribute is referenced. Otherwise it points to a
    specific element of the attribute. For example and entry in a buffer.
    """

    cosem_attribute: CosemAttribute
    data_index: int = attr.ib(default=0)

    @classmethod
    def from_bytes(cls, source_bytes) -> "CaptureObject":
        """
        It should be a structure of 4 elements-
        """
        # data = utils.parse_as_dlms_data(source_bytes)
        raise NotImplementedError()

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.extend(b"\x02\x04")  # A structure of 4 elements
        out.extend(
            dlms_data.UnsignedLongData(self.cosem_attribute.interface.value).to_bytes()
        )
        out.extend(
            dlms_data.OctetStringData(
                self.cosem_attribute.instance.to_bytes()
            ).to_bytes()
        )
        out.extend(dlms_data.IntegerData(self.cosem_attribute.attribute).to_bytes())
        out.extend(dlms_data.UnsignedLongData(self.data_index).to_bytes())
        return bytes(out)



import pytest

from dlms_cosem import cosem


class TestObis:
    def test_obis_to_dotted(self):
        obis = cosem.Obis(1, 0, 1, 8, 0, 255)
        assert obis.to_string(separator=".") == "1.0.1.8.0.255"

    def test_obis_from_bytes(self):
        data = b"\x00\x00+\x01\x00\xff"
        assert cosem.Obis.from_bytes(data) == cosem.Obis(0, 0, 43, 1, 0, 255)

    def test_obis_to_bytes(self):
        data = b"\x00\x00+\x01\x00\xff"
        assert cosem.Obis(0, 0, 43, 1, 0, 255).to_bytes() == data

    @pytest.mark.parametrize(
        "test_input,expected",
        [
            ("1-0:1.8.0.255", cosem.Obis(1, 0, 1, 8, 0, 255)),
            ("1-0:1.8.0", cosem.Obis(1, 0, 1, 8, 0, 255)),
            ("1-0-1-8-0-255", cosem.Obis(1, 0, 1, 8, 0, 255)),
            ("1-0-1-8-0", cosem.Obis(1, 0, 1, 8, 0, 255)),
            ("1.0.1.8.0.255", cosem.Obis(1, 0, 1, 8, 0, 255)),
            ("1.0.1.8.0", cosem.Obis(1, 0, 1, 8, 0, 255)),
        ],
    )
    def test_obis_from_string(self, test_input: str, expected: cosem.Obis):
        assert cosem.Obis.from_string(test_input) == expected

    def test_to_string(self):
        assert cosem.Obis.from_string("1-0:1.8.0.255").to_string() == "1-0:1.8.0.255"

    def test_non_parsable_raises_value_error(self):
        with pytest.raises(ValueError):
            cosem.Obis.from_string("1.8.0")



class LocalProtocolError(Exception):
    """Error in HDLC Protocol"""


class HdlcException(Exception):
    """Base class for HDLC protocol parts"""


class HdlcParsingError(HdlcException):
    """An error occurred then parsing bytes into HDLC object"""


class MissingHdlcFlags(HdlcParsingError):
    """Frame is not enclosed byt HDLC flags"""



import pytest

from dlms_cosem.hdlc import address, fields, frames
from dlms_cosem.hdlc.address import HdlcAddress
from dlms_cosem.io import HdlcTransport


def test_hdlc_frame_format_field_from_bytes():
    in_bytes = bytes.fromhex("a01d")
    f = fields.DlmsHdlcFrameFormatField.from_bytes(in_bytes)
    assert f.length == 29
    assert not f.segmented


def test_hdlc_frame_format_filed_from_bytes_segmented():
    in_bytes = bytes.fromhex("a81d")
    f = fields.DlmsHdlcFrameFormatField.from_bytes(in_bytes)
    assert f.length == 29
    assert f.segmented


def test_hdlc_frame_format_raises_value_error_too_long():
    with pytest.raises(ValueError):
        fields.DlmsHdlcFrameFormatField(length=999999, segmented=False)


def test_hdlc_frame_format_to_bytes_not_segmented():
    f = fields.DlmsHdlcFrameFormatField(length=29, segmented=False)
    assert f.to_bytes().hex() == "a01d"


def test_hdlc_frame_format_to_bytes_segmented():
    f = fields.DlmsHdlcFrameFormatField(length=29, segmented=True)
    assert f.to_bytes().hex() == "a81d"


class TestHdlcAddress:
    def test_find_address(self):
        frame = "7ea87e210396a4090f01160002020f02160002020f03160002020f04160002020f05160002020f0616000204120017110009060000160000ff0202010902030f0116010002030f0216010002030f0316010002030f0416010002030f0516010002030f0616010002030f0716010002030f0816010002030f09160100016ff37e"
        frame_bytes = bytes.fromhex(frame)
        (
            destination_address,
            source_address,
        ) = address.HdlcAddress.find_address_in_frame_bytes(frame_bytes)
        assert destination_address == (16, None, 1)
        assert source_address == (1, None, 1)

    @pytest.mark.parametrize(
        "hdlc_address,resulting_bytes",
        [
            ((1, None, "client"), b"\x03"),
            ((16, None, "client"), b"\x21"),
            ((0b1001010, None, "server"), b"\x95"),
            ((0b0100101, None, "client"), b"\x4b"),
            ((1, 17, "server"), b"\x02\x23"),
        ],
        # TODO: need to find references of multiy byte addresses to test.
    )
    def test_client_address_just_logical(self, hdlc_address, resulting_bytes):
        add = address.HdlcAddress(
            logical_address=hdlc_address[0],
            physical_address=hdlc_address[1],
            address_type=hdlc_address[2],
        )
        assert add.to_bytes() == resulting_bytes

    def test_find_4_byte_address(self):
        """
        Source address is using 4 bytes.
        :return:
        """
        frame1 = b"~\xa0#\x03\x00\x02\x00#s\xc0H\x81\x80\x14\x05\x02\x00\x80\x06\x02\x00\x80\x07\x04\x00\x00\x00\x01\x08\x04\x00\x00\x00\x01\xcej~"
        frame2 = b"~\xa0#!\x00\x02\x00#s\xf6\xc5\x81\x80\x14\x05\x02\x00\x80\x06\x02\x00\x80\x07\x04\x00\x00\x00\x01\x08\x04\x00\x00\x00\x01\xcej~"

        (
            destination_address,
            source_address,
        ) = address.HdlcAddress.find_address_in_frame_bytes(frame1)
        assert destination_address == (1, None, 1)
        assert source_address == (1, 17, 4)

        (
            destination_address,
            source_address,
        ) = address.HdlcAddress.find_address_in_frame_bytes(frame2)
        assert destination_address == (16, None, 1)
        assert source_address == (1, 17, 4)


class TestCrc:
    def test_crc(self):
        data = "033f"
        correct_crc = "5bec"

        _crc = frames.HCS
        result = _crc.calculate_for(bytes.fromhex(data))
        assert result == bytes.fromhex(correct_crc)


class TestHdlcFrameValidation:
    def test_frame_is_enclosed_by_hdlc_flag(self):
        data = b"\x7effff\x7e"
        assert frames.frame_is_enclosed_by_hdlc_flags(data)

    @pytest.mark.parametrize("data", [b"\x7effff", b"ffff\x7e", b"ffff"])
    def test_frame_is_enclosed_by_hdlc_flag_fail(self, data):
        assert not frames.frame_is_enclosed_by_hdlc_flags(data)


class TestSnrmFrame:
    def test_parses_correct(self):
        out_data = bytes.fromhex("7ea00802232193bd647e")
        destination_address = address.HdlcAddress(
            logical_address=1, physical_address=17, address_type="server"
        )

        # Public client
        source_address = address.HdlcAddress(
            logical_address=16, physical_address=None, address_type="client"
        )
        snrm = frames.SetNormalResponseModeFrame(destination_address, source_address)

        assert snrm.to_bytes() == out_data


class TestUAFrame:
    def test_parser_correctly(self):
        out_data = bytes.fromhex(
            "7EA01F21022373E6C781801205019A06019A070400000001080400000001CCA27E"
        )
        # from device so public client is destination address.
        source_address = address.HdlcAddress(
            logical_address=1, physical_address=17, address_type="server"
        )

        # Public client
        destination_address = address.HdlcAddress(
            logical_address=16, physical_address=None, address_type="client"
        )

        information = out_data[9:-3]
        ua = frames.UnNumberedAcknowledgmentFrame(
            destination_address, source_address, information
        )
        print(ua.hcs.hex())
        print(ua.to_bytes().hex())
        print(out_data.hex())
        assert ua.to_bytes() == out_data

    def test_from_bytes(self):
        in_data = b"~\xa0\x1f!\x02#s\xe6\xc7\x81\x80\x12\x05\x01\x9a\x06\x01\x9a\x07\x04\x00\x00\x00\x01\x08\x04\x00\x00\x00\x01\xcc\xa2~"
        frame = frames.UnNumberedAcknowledgmentFrame.from_bytes(in_data)
        assert in_data == frame.to_bytes()

    def test_4_byte_address(self):
        frame1 = b"~\xa0#\x03\x00\x02\x00#s\xc0H\x81\x80\x14\x05\x02\x00\x80\x06\x02\x00\x80\x07\x04\x00\x00\x00\x01\x08\x04\x00\x00\x00\x01\xcej~"
        frame2 = b"~\xa0#!\x00\x02\x00#s\xf6\xc5\x81\x80\x14\x05\x02\x00\x80\x06\x02\x00\x80\x07\x04\x00\x00\x00\x01\x08\x04\x00\x00\x00\x01\xcej~"

        frame = frames.UnNumberedAcknowledgmentFrame.from_bytes(frame1)
        assert frame1 == frame.to_bytes()

        frame = frames.UnNumberedAcknowledgmentFrame.from_bytes(frame2)
        assert frame2 == frame.to_bytes()


class TestInformationFrame:
    def test_construct(self):
        total = bytes.fromhex(
            "7EA02C02232110AF9FE6E600601DA109060760857405080101BE10040E01000000065F1F0400001E1DFFFFC5E47E"
        )
        information_part = bytes.fromhex(
            "E6E600601DA109060760857405080101BE10040E01000000065F1F0400001E1DFFFF"
        )
        server_address = address.HdlcAddress(
            logical_address=1, physical_address=17, address_type="server"
        )

        # Public client
        client_address = address.HdlcAddress(
            logical_address=16, physical_address=None, address_type="client"
        )
        frame = frames.InformationFrame(
            destination_address=server_address,
            source_address=client_address,
            payload=information_part,
            send_sequence_number=0,
            receive_sequence_number=0,
            segmented=False,
            final=True,
        )

        print(total)
        print(frame.to_bytes())
        assert frame.to_bytes() == total


class TestInformationResponseFrame:
    def test_contruct(self):
        total = bytes.fromhex(
            "7EA0382102233034E7E6E7006129A109060760857405080101A203020100A305A103020100BE10040E0800065F1F0400001E1D04C80007B86A7E"
        )
        information_part = bytes.fromhex(
            "E6E7006129A109060760857405080101A203020100A305A103020100BE10040E0800065F1F0400001E1D04C80007"
        )
        server_address = address.HdlcAddress(
            logical_address=1, physical_address=17, address_type="server"
        )

        # Public client
        client_address = address.HdlcAddress(
            logical_address=16, physical_address=None, address_type="client"
        )
        frame = frames.InformationFrame(
            destination_address=client_address,
            source_address=server_address,
            payload=information_part,
            send_sequence_number=0,
            receive_sequence_number=1,
            segmented=False,
            final=True,
        )

        print(f"length = {frame.frame_length}  :: {len(frame.to_bytes())}")

        print(total.hex())
        print(frame.to_bytes().hex())
        assert frame.to_bytes() == total

    def test_rebuild_frame(self):
        in_data = b"~\xa08!\x02#04\xe7\xe6\xe7\x00a)\xa1\t\x06\x07`\x85t\x05\x08\x01\x01\xa2\x03\x02\x01\x00\xa3\x05\xa1\x03\x02\x01\x00\xbe\x10\x04\x0e\x08\x00\x06_\x1f\x04\x00\x00\x1e\x1d\x04\xc8\x00\x07\xb8j~"
        info = frames.InformationFrame.from_bytes(in_data)
        print(info)
        assert info.to_bytes().hex() == in_data.hex()


class TestInformationControlField:
    def test_from_bytes(self):
        in_byte = bytes.fromhex("30")
        ctrl = fields.InformationControlField.from_bytes(in_byte)
        assert ctrl.receive_sequence_number == 1
        assert ctrl.send_sequence_number == 0
        assert ctrl.final


class TestUnnumberedInformationFrame:
    def test_parse(self):
        data = b"~\xa2CA\x08\x83\x13\x85\xeb\xe6\xe7\x00\x0f@\x00\x00\x00\x00\x01\x1b\x02\x02\t\x06\x00\x00\x01\x00\x00\xff\t\x0c\x07\xe3\x0c\x10\x01\x07;(\xff\x80\x00\xff\x02\x03\t\x06\x01\x00\x01\x07\x00\xff\x06\x00\x00\x04b\x02\x02\x0f\x00\x16\x1b\x02\x03\t\x06\x01\x00\x02\x07\x00\xff\x06\x00\x00\x00\x00\x02\x02\x0f\x00\x16\x1b\x02\x03\t\x06\x01\x00\x03\x07\x00\xff\x06\x00\x00\x05\xe3\x02\x02\x0f\x00\x16\x1d\x02\x03\t\x06\x01\x00\x04\x07\x00\xff\x06\x00\x00\x00\x00\x02\x02\x0f\x00\x16\x1d\x02\x03\t\x06\x01\x00\x1f\x07\x00\xff\x10\x00\x00\x02\x02\x0f\xff\x16!\x02\x03\t\x06\x01\x003\x07\x00\xff\x10\x00K\x02\x02\x0f\xff\x16!\x02\x03\t\x06\x01\x00G\x07\x00\xff\x10\x00\x00\x02\x02\x0f\xff\x16!\x02\x03\t\x06\x01\x00 \x07\x00\xff\x12\t\x03\x02\x02\x0f\xff\x16#\x02\x03\t\x06\x01\x004\x07\x00\xff\x12\t\xc3\x02\x02\x0f\xff\x16#\x02\x03\t\x06\x01\x00H\x07\x00\xff\x12\t\x04\x02\x02\x0f\xff\x16#\x02\x03\t\x06\x01\x00\x15\x07\x00\xff\x06\x00\x00\x00\x00\x02\x02\x0f\x00\x16\x1b\x02\x03\t\x06\x01\x00\x16\x07\x00\xff\x06\x00\x00\x00\x00\x02\x02\x0f\x00\x16\x1b\x02\x03\t\x06\x01\x00\x17\x07\x00\xff\x06\x00\x00\x00\x00\x02\x02\x0f\x00\x16\x1d\x02\x03\t\x06\x01\x00\x18\x07\x00\xff\x06\x00\x00\x00\x00\x02\x02\x0f\x00\x16\x1d\x02\x03\t\x06\x01\x00)\x07\x00\xff\x06\x00\x00\x04b\x02\x02\x0f\x00\x16\x1b\x02\x03\t\x06\x01\x00*\x07\x00\xff\x06\x00\x00\x00\x00\x02\x02\x0f\x00\x16\x1b\x02\x03\t\x06\x01\x00+\x07\x00\xff\x06\x00\x00\x05\xe2\x02\x02\x0f\x00\x16\x1d\x02\x03\t\x06\x01\x00,\x07\x00\xff\x06\x00\x00\x00\x00\x02\x02\x0f\x00\x16\x1d\x02\x03\t\x06\x01\x00=\x07\x00\xff\x06\x00\x00\x00\x00\x02\x02\x0f\x00\x16\x1b\x02\x03\t\x06\x01\x00>\x07\x00\xff\x06\x00\x00\x00\x00\x02\x02\x0f\x00\x16\x1b\x02\x03\t\x06\x01\x00?\x07\x00\xff\x06\x00\x00\x00\x00\x02\x02\x0f\x00\x16\x1d\x02\x03\t\x06\x01\x00@\x07\x00\xff\x06\x00\x00\x00\x00\x02\x02\x0f\x00\x16\x1d\x02\x03\t\x06\x01\x00\x01\x08\x00\xff\x06\x00\x99Y\x86\x02\x02\x0f\x00\x16\x1e\x02\x03\t\x06\x01\x00\x02\x08\x00\xff\x06\x00\x00\x00\x08\x02\x02\x0f\x00\x16\x1e\x02\x03\t\x06\x01\x00\x03\x08\x00\xff\x06\x00d\xedK\x02\x02\x0f\x00\x16 \x02\x03\t\x06\x01\x00\x04\x08\x00\xff\x06\x00\x00\x00\x05\x02\x02\x0f\x00\x16 \xbe@~"
        ui = frames.UnnumberedInformationFrame.from_bytes(data)
        assert ui.to_bytes() == data


class TestUnnumberedInformationControlField:
    def test_parse(self):
        data = b"\x13"
        cf = fields.UnnumberedInformationControlField.from_bytes(data)
        assert cf.to_bytes() == data


class TestFiorentiniLocal:
    """ no physical address when talking to fiorentini meters"""

    def test_frame(self):
        response = b'~\xa0\x07!\x03s\x01@~'

        frame = frames.UnNumberedAcknowledgmentFrame.from_bytes(response)

    def test_ua_frame(self):
        frame = frames.UnNumberedAcknowledgmentFrame(
            destination_address=HdlcAddress(logical_address=16, physical_address=None, address_type='client',
                                            extended_addressing=False),
            source_address=HdlcAddress(logical_address=1, physical_address=None, address_type='server',
                                       extended_addressing=False), payload=b'', segmented=False, final=True)

        fsc = frame.fcs
    def test_fcs(self):
        data = b'~\xa0\x07!\x03s\x01@~'
        fcs = data[-3:-1]
        assert fcs == b"\x01@"
        # frame_content should be everything except flag and fcs


# class TestKaifaMeter:
#
#     def test_kaifa_data(self):
#         # data = "7ea09b01000110561be6e7000f40000000090c07e7090401103400ff800000021209074b464d5f30303109103733343031353730313132353335343409084d41333034483444060000044f0600000000060000000006000000c0060000088f06000005aa060000057c06000008da06000008f906000008e6090c07e7090401103400ff8000000608c141c9060000000006001ae03806013151959d787e"
#         data = "7ea11d01000110b0aee6e7000f4000000000022409060100000281ff09074b464d5f30303109060000600100ff09103733343031353730333037383433393509060000600107ff09074d41333034483409060100010700ff060000017209060100020700ff060000000009060100030700ff060000000009060100040700ff0600000050090601001f0700ff06000001be09060100330700ff060000047b09060100470700ff060000009409060100200700ff06000008ee09060100340700ff06000008e609060100480700ff06000008dd09060000010000ff090c07e709040111132dffffc40009060100010800ff0600bc0d9009060100020800ff060000000009060100030800ff06001c05dc09060100040800ff06000fed42acfd7e"
#         frame = frames.UnnumberedInformationFrame.from_bytes(bytes.fromhex(data))
#         print(frame)



import os
import sys
from shutil import rmtree

from setuptools import Command, find_packages, setup

# Package meta-data.
NAME = "dlms-cosem"
DESCRIPTION = "A Python library for DLMS/COSEM"
URL = "https://github.com/pwitab/dlms-cosem"
PROJECT_URLS = {
    "Documentation": "https://www.dlms.dev/",
    "Bug Tracker": "https://github.com/pwitab/dlms-cosem/issues",
    "Source Code": "https://github.com/pwitab/dlms-cosem",
}
EMAIL = "henrik@pwit.se"
AUTHOR = "Henrik Palmlund Wahlgren @ Palmlund Wahlgren Innovative Technology AB"
REQUIRES_PYTHON = "~=3.7"
VERSION = "25.1.0"

# What packages are required for this module to be executed?
REQUIRED = [
    "attrs>=22.2.0",
    "pyserial>=3.5",
    "cryptography>=35.0.0",
    "asn1crypto>=1.4.0",
    "python-dateutil>=2.8.1",
    "typing-extensions>=3.10",
    "structlog>=22.1.0",
]

DOC_PACKAGES = ["mkdocs", "mkdocs-material"]
TEST_PACKAGES = ["pytest", "pytest-cov", "pytest-sugar"]
DEV_PACKAGES = ["pre-commit"] + DOC_PACKAGES + TEST_PACKAGES

EXTRAS = {
    "docs": DOC_PACKAGES,
    "test": TEST_PACKAGES,
    "dev": DEV_PACKAGES,
}

CLASSIFIERS = [
    "Intended Audience :: Developers",
    "Natural Language :: English",
    "License :: Other/Proprietary License",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: Implementation :: CPython",
    "Topic :: Software Development :: Libraries :: Python Modules",
    "Topic :: Communications",

]

here = os.path.abspath(os.path.dirname(__file__))


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel".format(sys.executable))

        self.status("Uploading the package to PyPI via Twine…")
        os.system("twine upload dist/*")

        self.status("Pushing git tags…")
        # os.system('git tag v{0}'.format(about['__version__']))
        os.system("git push --tags")

        sys.exit()


with open("README.md") as readme_file:
    readme = readme_file.read()

with open("HISTORY.md") as history_file:
    history = history_file.read()

setup(
    name=NAME,
    version=VERSION,
    python_requires=REQUIRES_PYTHON,
    description=DESCRIPTION,
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    maintainer=AUTHOR,
    maintainer_email=EMAIL,
    url=URL,
    project_urls=PROJECT_URLS,
    packages=find_packages(exclude=("tests",)),
    entry_points={},
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license="MIT",
    zip_safe=False,
    keywords="AMR, Metering, smart meters, MDM, dlms, cosem",
    classifiers=CLASSIFIERS,
    # $ setup.py publish support.
    cmdclass={"upload": UploadCommand},
)



import abc

import attr

from dlms_cosem.hdlc import exceptions as hdlc_exceptions
from dlms_cosem.hdlc import validators


class _AbstractHdlcControlField(abc.ABC):
    """
    Control field is represented by 1 bytes of data.

    Indicates the type of commands or responses, and contains HDLC
    sequence numbers, where appropriate. The last bits of the control field (CTRL)
    identify the type of the HDLC frame

    The is_final represents the poll/final bit. Indicates if the frame is the last one
    of a secquence. Setting the bit releases the control to the other party.
    """

    @property
    @abc.abstractmethod
    def is_final(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def to_bytes(self) -> bytes:
        raise NotImplementedError()


@attr.s(auto_attribs=True)
class SnrmControlField(_AbstractHdlcControlField):
    """
    S-frame fo SNRM request.
    """

    def is_final(self):
        """ 'Almost' all the time a SNRM frame is contaned in single frame."""
        # TODO: Handle multi frame
        return True

    def to_bytes(self) -> bytes:
        out = 0b10000011
        if self.is_final:
            out |= 0b00010000
        return out.to_bytes(1, "big")


@attr.s(auto_attribs=True)
class UaControlField(_AbstractHdlcControlField):
    """
    S-frame for Unacknowladge Answer.
    """

    def is_final(self):
        """
        Most UA is only one frame. But in the HDLC setup it can be longer depending on
        the data send.
        # TODO: Handle multi frame
        """
        return True

    def to_bytes(self) -> bytes:
        """
        Returns byte representation of the field.
        """
        out = 0b01100011
        if self.is_final:
            out |= 0b00010000
        return out.to_bytes(1, "big")


@attr.s(auto_attribs=True)
class DisconnectControlField(_AbstractHdlcControlField):
    """
    S-frame for disconnect.
    """

    def is_final(self):
        """
        Always final
        """
        return True

    def to_bytes(self) -> bytes:
        """
        Returns byte representation of the field.
        """
        out = 0b01000011
        if self.is_final:
            out |= 0b00010000
        return out.to_bytes(1, "big")


@attr.s(auto_attribs=True)
class ReceiveReadyControlField(_AbstractHdlcControlField):
    """
    RR-frame for ack.
    """

    receive_sequence_number: int = attr.ib(
        validator=[validators.validate_information_sequence_number]
    )

    def is_final(self):
        """
        Always final
        """
        return True

    def to_bytes(self) -> bytes:
        """
        Returns byte representation of the field.
        """
        out = 0b00000001
        out += self.receive_sequence_number << 5
        if self.is_final:
            out |= 0b00010000
        return out.to_bytes(1, "big")

    @classmethod
    def from_bytes(cls, in_byte: bytes):
        if len(in_byte) != 1:
            raise ValueError(
                f"ReceiveReadyControlField can only be 1 bytes. Got {len(in_byte)}"
            )
        value = int.from_bytes(in_byte, "big")
        control_frame = bool(value & 0b00000001)
        if not control_frame:
            raise ValueError("Frame is an information frame not a ReceiveReadyFrame")
        rsn = (value & 0b11100000) >> 5
        return cls(rsn)


@attr.s(auto_attribs=True)
class InformationControlField(_AbstractHdlcControlField):
    """
    Information control field also contains information about the acknowlegde frames
    sent between the client and server.

    The `send_sequence_number` hold information about the enumeration of the current
    frame in transit.

    The `receive_sequence_number` holds information about the enumeration of the next
    frame the sender is expecting to be delivered. This can also be used as a frame
    acknowledgement. If a sender doesn't expects the next frame the other part is
    about to send it has not received the last frame.

    `send_sequence_number` and `receive_sequence_number` are in DLMS limited to 3 bits
    and can take the value of 0-7. If a communication exceeds 8 consecutive frames
    from a part the number rolls over from 7 to 0.

    `final` shows if the frame is the last one that is to be sent by the sender and
    sending control can be relased to the other party.

    information control field indicate the it is part of an information field by
    setting the LSB (bit 0) to 0.

    `send_sequence_number` is encoded in bit 1-3, `receive_sequence_number` is encoded
    in bit 5-7 and the final flag is encoded in bit 4.


    """

    send_sequence_number: int = attr.ib(
        validator=[validators.validate_information_sequence_number]
    )
    receive_sequence_number: int = attr.ib(
        validator=[validators.validate_information_sequence_number]
    )
    final: bool = attr.ib(default=True)

    @property
    def is_final(self):
        return self.final

    def to_bytes(self) -> bytes:
        out = 0b00000000
        out += self.send_sequence_number << 1
        out += self.receive_sequence_number << 5
        if self.is_final:
            out |= 0b00010000
        return out.to_bytes(1, "big")

    @classmethod
    def from_bytes(cls, in_byte: bytes):
        if len(in_byte) != 1:
            raise ValueError(
                f"InformationControlField can only be 1 bytes. Got {len(in_byte)}"
            )
        value = int.from_bytes(in_byte, "big")
        not_info_frame = bool(value & 0b00000001)
        if not_info_frame:
            raise ValueError(
                "Byte is not representing a InformationControlField. LSB is 1, should be 0"
            )
        ssn = (value & 0b00001110) >> 1

        rsn = (value & 0b11100000) >> 5
        final = bool(value & 0b00010000)
        return cls(ssn, rsn, final)


@attr.s(auto_attribs=True)
class UnnumberedInformationControlField(_AbstractHdlcControlField):
    """
    Used for UnnumberedInformationFrames.
    """

    final: bool = attr.ib(default=True)

    @property
    def is_final(self):
        return self.final

    def to_bytes(self) -> bytes:
        out = 0b00000011
        if self.is_final:
            out |= 0b00010000
        return out.to_bytes(1, "big")

    @classmethod
    def from_bytes(cls, in_byte: bytes):
        if len(in_byte) != 1:
            raise ValueError(
                f"InformationControlField can only be 1 bytes. Got {len(in_byte)}"
            )
        value = int.from_bytes(in_byte, "big")
        is_unnumbered_info_frame = bool(value & 0b00000011)
        if not is_unnumbered_info_frame:
            raise ValueError(
                "Byte is not representing a UnnumberedInformationControlField."
            )
        final = bool(value & 0b00010000)
        return cls(final)


def validate_frame_length(instance, attribute, value):
    if value > 0b11111111111:
        raise ValueError("frame length is to long")
    if value < 0:
        raise ValueError("frame length cannot be negative")


@attr.s(auto_attribs=True)
class DlmsHdlcFrameFormatField:
    """
    2 bytes.

    The 4 leftmost bits represents the HDLC frame format.
    DLMS used HDLC frame format 3 -> 0b1010 -> 0xA. This is always the same in all
    frames

    The bit 11 is the segmentation bit. If set it indicates that the
    data the frame consists of is not complete and has been segmented into several
    frames.

    The bit 0-10 rightmost bits represents the frame length.
    Lenght of a frame is calculated excluding the frame tags (0x7e).

    """

    length: int = attr.ib(validator=[validate_frame_length])
    segmented: bool

    @classmethod
    def from_bytes(cls, in_bytes: bytes):
        if len(in_bytes) != 2:
            raise hdlc_exceptions.HdlcParsingError(
                f"HDLC frame format length is {len(in_bytes)}, should be 2"
            )
        if not cls.correct_frame_format(in_bytes):
            raise hdlc_exceptions.HdlcParsingError(
                f"Received a HDLC frame of the incorrect format: {in_bytes!r}"
            )
        segmented = bool(in_bytes[0] & 0b00001000)
        length = cls.get_length_from_bytes(in_bytes)
        return cls(length, segmented)

    @staticmethod
    def correct_frame_format(_bytes: bytes) -> bool:
        leftmost: int = _bytes[0]
        masked_leftmost = leftmost & 0b11110000
        return masked_leftmost == 0b10100000

    @staticmethod
    def get_length_from_bytes(_bytes: bytes) -> int:
        """Length is in the rightmost 11 bits"""
        total = int.from_bytes(_bytes, "big")
        return total & 0b0000011111111111

    def to_bytes(self) -> bytes:
        total = 0b1010000000000000 | self.length
        if self.segmented:
            total = total | 0b0000100000000000
        return total.to_bytes(2, "big")



import abc
from typing import *

import attr

from dlms_cosem.crc import CRCCCITT
from dlms_cosem.hdlc import address
from dlms_cosem.hdlc import exceptions as hdlc_exceptions
from dlms_cosem.hdlc import fields, validators
from dlms_cosem.hdlc.address import HdlcAddress

HDLC_FLAG = b"\x7e"

FCS = CRCCCITT()
HCS = FCS


def frame_is_enclosed_by_hdlc_flags(frame_bytes: bytes):
    first: int = frame_bytes[0]
    last: int = frame_bytes[-1]
    if first != last:
        return False
    if first != ord(HDLC_FLAG) or last != ord(HDLC_FLAG):
        return False

    return True


def frame_has_correct_length(control_field_length: int, frame_bytes: bytes):
    # Control field length doesn't calculate HDLC flags.
    if (control_field_length + 2) != len(frame_bytes):
        return False
    return True


class _AbstractHdlcFrame(abc.ABC):
    """
    HDLC frames start and end with the HDLC Frame flag 0x7E

    Frame:
    Flag (1 byte), Format (2 bytes), Destination Address (1-4 bytes),
    Source Address (1-4 bytes), Control (1 byte), Header check sequence (2 bytes),
    Information (n bytes), Frame check sequence (2 bytes), Flag (1 byte)

    The header check sequence field is only present when the frame has a Information field.
    """

    @property
    @abc.abstractmethod
    def frame_length(self) -> int:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def hcs(self) -> bytes:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def fcs(self) -> bytes:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def information(self) -> bytes:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def header_content(self) -> bytes:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def frame_content(self) -> bytes:
        raise NotImplementedError()

    @abc.abstractmethod
    def to_bytes(self) -> bytes:
        raise NotImplementedError()


@attr.s(auto_attribs=True)
class BaseHdlcFrame(_AbstractHdlcFrame):
    """
    Base class for HDLC frames and holds general behavior
    """

    destination_address: HdlcAddress
    source_address: HdlcAddress
    payload: Optional[bytes] = attr.ib(default=None)
    segmented: bool = attr.ib(default=False)
    final: bool = attr.ib(default=True)

    fixed_length_bytes: ClassVar = 7

    @property
    def frame_length(self) -> int:
        return (
            self.fixed_length_bytes
            + self.destination_address.length
            + self.source_address.length
            + len(self.information)
        )

    @property
    def hcs(self) -> bytes:
        return HCS.calculate_for(self.header_content)

    @property
    def fcs(self) -> bytes:
        return FCS.calculate_for(self.frame_content)

    @property
    def information(self) -> bytes:
        raise NotImplementedError()

    @property
    def header_content(self) -> bytes:

        return b"".join(
            [
                fields.DlmsHdlcFrameFormatField(
                    length=self.frame_length, segmented=self.segmented
                ).to_bytes(),
                self.destination_address.to_bytes(),
                self.source_address.to_bytes(),
                self.get_control_field().to_bytes(),
            ]
        )

    @property
    def frame_content(self) -> bytes:
        return b"".join([self.header_content, self.hcs, self.information])

    def to_bytes(self):
        return b"".join([HDLC_FLAG, self.frame_content, self.fcs, HDLC_FLAG])

    def get_control_field(self):
        """
        Should return the correct control field of the frame.
        """
        raise NotImplementedError()

    @staticmethod
    def extract_format_field_from_bytes(
        frame_bytes: bytes,
    ) -> fields.DlmsHdlcFrameFormatField:
        """
        The format field is always on the same place in all frames.
        It is the first 2 bytest after the HDLC Flag.
        """
        return fields.DlmsHdlcFrameFormatField.from_bytes(frame_bytes[1:3])


@attr.s(auto_attribs=True)
class SetNormalResponseModeFrame(BaseHdlcFrame):
    """
    SetNormalResponseMode Frame (SNRM-frame) is used to start a new HDLC connection.
    """

    fixed_length_bytes = 5

    @property
    def hcs(self) -> bytes:
        """
        SetNormalResponseModeFrame is an HDLC S-frame and does not contain an
        information field. That means that there is no HCS field present, only FCS
        # TODO: is it present when using the information field as parameter field for
            negotiating HDLC parameters?
        """
        return b""

    @property
    def information(self) -> bytes:
        """
        The information field on an SNRM request can be used to negotiate HDLC
        connection parameters, (window size, max_frame_length etc.) It is not supported.

        By not sending any information default values will be assumed.
        Window size = 1, max transmit size = 128 bytes.
        """
        return b""

    def get_control_field(self):
        return fields.SnrmControlField()


@attr.s(auto_attribs=True)
class UnNumberedAcknowledgmentFrame(BaseHdlcFrame):

    fixed_length_bytes = 7

    @property
    def frame_length(self) -> int:

        if self.information:
            # with an informatoin field it should have HSC
            fixed = self.fixed_length_bytes
        else:
            fixed = self.fixed_length_bytes - 2

        return (
                fixed
                + self.destination_address.length
                + self.source_address.length
                + len(self.information)
        )


    @property
    def information(self) -> bytes:
        """
        Information field on UA does not contain an LLC as it is not a true
        information field.
        """
        out: List[bytes] = list()
        if self.payload:
            out.append(self.payload)

        return b"".join(out)

    @property
    def hcs(self) -> bytes:
        """
        UnNumberedAcknowledgmentFrame is an HDLC S-frame and does if it does not contain an
        information field it should also not contain a HCS field. That means that there is no HCS field present, only FCS
        """
        if self.payload:
            return HCS.calculate_for(self.header_content)
        else:
            return b""

    def get_control_field(self):
        return fields.UaControlField()

    @classmethod
    def from_bytes(cls, frame_bytes: bytes):
        if not frame_is_enclosed_by_hdlc_flags(frame_bytes):
            raise hdlc_exceptions.MissingHdlcFlags()

        frame_format = BaseHdlcFrame.extract_format_field_from_bytes(frame_bytes)

        if not frame_has_correct_length(frame_format.length, frame_bytes):
            raise hdlc_exceptions.HdlcParsingError(
                f"Frame data is not of length specified in frame format field. "
                f"Should be {frame_format.length} but is {len(frame_bytes)}"
            )

        destination_address = address.HdlcAddress.destination_from_bytes(
            frame_bytes, "client"
        )
        source_address = address.HdlcAddress.source_from_bytes(frame_bytes, "server")

        hcs_position = 1 + 2 + destination_address.length + source_address.length + 1
        hcs = frame_bytes[hcs_position : hcs_position + 2]
        fcs = frame_bytes[-3:-1]

        information = frame_bytes[hcs_position + 2 : -3]

        frame = cls(destination_address, source_address, information)

        if frame.hcs:
            "Some frames might not have hcs so we should not check it."
            if hcs != frame.hcs:
                raise hdlc_exceptions.HdlcParsingError(
                    f"HCS is not correct. " f"Calculated: {frame.hcs!r}, in data: {hcs!r}"
                )

        if fcs != frame.fcs:
            raise hdlc_exceptions.HdlcParsingError("FCS is not correct")

        return frame


@attr.s(auto_attribs=True)
class ReceiveReadyFrame(BaseHdlcFrame):
    fixed_length_bytes = 5

    receive_sequence_number: int = attr.ib(
        validator=[validators.validate_information_sequence_number], default=0
    )

    @property
    def hcs(self) -> bytes:
        """No information field in the frame so no hcs. Only FCS"""
        return b""

    @property
    def information(self) -> bytes:
        """
        No information field present
        """
        return b""

    def get_control_field(self):
        return fields.ReceiveReadyControlField(
            receive_sequence_number=self.receive_sequence_number
        )

    @classmethod
    def from_bytes(cls, frame_bytes: bytes):
        if not frame_is_enclosed_by_hdlc_flags(frame_bytes):
            raise hdlc_exceptions.MissingHdlcFlags()

        frame_format = BaseHdlcFrame.extract_format_field_from_bytes(frame_bytes)

        if not frame_has_correct_length(frame_format.length, frame_bytes):
            raise hdlc_exceptions.HdlcParsingError(
                f"Frame data is not of length specified in frame format field. "
                f"Should be {frame_format.length} but is {len(frame_bytes)}"
            )

        destination_address = address.HdlcAddress.destination_from_bytes(
            frame_bytes, "client"
        )
        source_address = address.HdlcAddress.source_from_bytes(frame_bytes, "server")
        control_byte_position = (
            1 + 2 + destination_address.length + source_address.length
        )
        control_byte = frame_bytes[control_byte_position : control_byte_position + 1]
        control = fields.ReceiveReadyControlField.from_bytes(control_byte)
        fcs = frame_bytes[-3:-1]

        frame = cls(
            destination_address=destination_address,
            source_address=source_address,
            receive_sequence_number=control.receive_sequence_number,
            final=control.is_final,
        )

        if fcs != frame.fcs:
            raise hdlc_exceptions.HdlcParsingError("FCS is not correct")

        return frame


@attr.s(auto_attribs=True)
class InformationFrame(BaseHdlcFrame):

    fixed_length_bytes = 7

    send_sequence_number: int = attr.ib(
        validator=[validators.validate_information_sequence_number], default=0
    )
    receive_sequence_number: int = attr.ib(
        validator=[validators.validate_information_sequence_number], default=0
    )

    @property
    def information(self) -> bytes:
        """
        Information request uses the LLC_COMMAND_HEADER
        """
        out_data: List[bytes] = list()
        if self.payload:
            out_data.append(self.payload)

        return b"".join(out_data)

    def get_control_field(self):
        return fields.InformationControlField(
            self.send_sequence_number, self.receive_sequence_number, self.final
        )

    @classmethod
    def from_bytes(cls, frame_bytes: bytes):

        if not frame_is_enclosed_by_hdlc_flags(frame_bytes):
            raise hdlc_exceptions.MissingHdlcFlags()

        frame_format = BaseHdlcFrame.extract_format_field_from_bytes(frame_bytes)

        if not frame_has_correct_length(frame_format.length, frame_bytes):
            raise hdlc_exceptions.HdlcParsingError(
                f"Frame data is not of length specified in frame format field. "
                f"Should be {frame_format.length} but is {len(frame_bytes)}"
            )

        destination_address = address.HdlcAddress.destination_from_bytes(
            frame_bytes, "client"
        )
        source_address = address.HdlcAddress.source_from_bytes(frame_bytes, "server")

        information_control_byte_position = (
            1 + 2 + destination_address.length + source_address.length
        )
        information_control_byte = frame_bytes[
            information_control_byte_position : information_control_byte_position + 1
        ]
        information_control = fields.InformationControlField.from_bytes(
            information_control_byte
        )

        hcs_position = 1 + 2 + destination_address.length + source_address.length + 1
        hcs = frame_bytes[hcs_position : hcs_position + 2]
        fcs = frame_bytes[-3:-1]
        information = frame_bytes[hcs_position + 2 : -3]

        frame = cls(
            destination_address,
            source_address,
            information,
            send_sequence_number=information_control.send_sequence_number,
            receive_sequence_number=information_control.receive_sequence_number,
            segmented=frame_format.segmented,
            final=information_control.final,
        )

        if hcs != frame.hcs:
            raise hdlc_exceptions.HdlcParsingError(
                f"HCS is not correct Calculated: {frame.hcs!r}, in data: {hcs!r}"
            )

        if fcs != frame.fcs:
            raise hdlc_exceptions.HdlcParsingError(
                f"FCS is not correct, Calculated: {frame.fcs!r}, in data: {fcs!r}"
            )

        return frame


@attr.s(auto_attribs=True)
class DisconnectFrame(BaseHdlcFrame):

    fixed_length_bytes = 5

    @property
    def hcs(self) -> bytes:
        """No information field in the frame so no hcs. Only FCS"""
        return b""

    @property
    def information(self) -> bytes:
        """
        No information field present
        """
        return b""

    def get_control_field(self):
        return fields.DisconnectControlField()

    @classmethod
    def from_bytes(cls, frame_bytes: bytes):
        if not frame_is_enclosed_by_hdlc_flags(frame_bytes):
            raise hdlc_exceptions.MissingHdlcFlags()

        frame_format = BaseHdlcFrame.extract_format_field_from_bytes(frame_bytes)

        if not frame_has_correct_length(frame_format.length, frame_bytes):
            raise hdlc_exceptions.HdlcParsingError(
                f"Frame data is not of length specified in frame format field. "
                f"Should be {frame_format.length} but is {len(frame_bytes)}"
            )

        destination_address = address.HdlcAddress.destination_from_bytes(
            frame_bytes, "server"
        )

        source_address = address.HdlcAddress.source_from_bytes(frame_bytes, "client")

        fcs = frame_bytes[-3:-1]

        frame = cls(destination_address, source_address)

        if fcs != frame.fcs:
            raise hdlc_exceptions.HdlcParsingError("FCS is not correct")

        return frame


@attr.s(auto_attribs=True)
class UnnumberedInformationFrame(BaseHdlcFrame):

    fixed_length_bytes = 7

    @property
    def information(self) -> bytes:
        """"""
        out_data: List[bytes] = list()
        if self.payload:
            out_data.append(self.payload)

        return b"".join(out_data)

    def get_control_field(self):
        return fields.UnnumberedInformationControlField(self.final)

    @classmethod
    def from_bytes(cls, frame_bytes: bytes):

        if not frame_is_enclosed_by_hdlc_flags(frame_bytes):
            raise hdlc_exceptions.MissingHdlcFlags()

        frame_format = BaseHdlcFrame.extract_format_field_from_bytes(frame_bytes)

        if not frame_has_correct_length(frame_format.length, frame_bytes):
            raise hdlc_exceptions.HdlcParsingError(
                f"Frame data is not of length specified in frame format field. "
                f"Should be {frame_format.length} but is {len(frame_bytes)}"
            )

        destination_address = address.HdlcAddress.destination_from_bytes(
            frame_bytes, "client"
        )
        source_address = address.HdlcAddress.source_from_bytes(frame_bytes, "server")

        information_control_byte_position = (
            1 + 2 + destination_address.length + source_address.length
        )
        information_control_byte = frame_bytes[
            information_control_byte_position : information_control_byte_position + 1
        ]
        information_control = fields.UnnumberedInformationControlField.from_bytes(
            information_control_byte
        )

        hcs_position = 1 + 2 + destination_address.length + source_address.length + 1
        hcs = frame_bytes[hcs_position : hcs_position + 2]
        fcs = frame_bytes[-3:-1]
        information = frame_bytes[hcs_position + 2 : -3]

        frame = cls(
            destination_address,
            source_address,
            information,
            segmented=frame_format.segmented,
            final=information_control.final,
        )

        if hcs != frame.hcs:
            raise hdlc_exceptions.HdlcParsingError(
                f"HCS is not correct Calculated: {frame.hcs!r}, in data: {hcs!r}"
            )

        if fcs != frame.fcs:
            raise hdlc_exceptions.HdlcParsingError(
                f"FCS is not correct, Calculated: {frame.fcs!r}, in data: {fcs!r}"
            )

        return frame



import pytest

from dlms_cosem import enumerations
from dlms_cosem.protocol.acse import ReleaseRequest


class TestDecodeRLRQ:
    def test_simple(self):
        data = bytes.fromhex("6203800100")  # Normal no user-information
        rlrq = ReleaseRequest.from_bytes(data)
        assert rlrq.reason == enumerations.ReleaseRequestReason.NORMAL
        assert rlrq.user_information is None
        assert data == rlrq.to_bytes()

    def test_with_ciphered_initiate_request(self):
        data = bytes.fromhex(
            "6239800100be34043221303001234567801302FF8A7874133D414CED25B42534D28DB0047720606B175BD52211BE6841DB204D39EE6FDB8E356855"
        )
        # No support for ciphnered adpus yet

        rlrq = ReleaseRequest.from_bytes(data)
        assert rlrq.reason == enumerations.ReleaseRequestReason.NORMAL



from dlms_cosem.protocol.acse.aare import ApplicationAssociationResponse
from dlms_cosem.protocol.acse.aarq import ApplicationAssociationRequest
from dlms_cosem.protocol.acse.base import *
from dlms_cosem.protocol.acse.rlre import ReleaseResponse
from dlms_cosem.protocol.acse.rlrq import ReleaseRequest
from dlms_cosem.protocol.acse.user_information import UserInformation

__all__ = [
    "ApplicationAssociationRequest",
    "ApplicationAssociationResponse",
    "ReleaseRequest",
    "ReleaseResponse",
    "AppContextName",
    "MechanismName",
    "UserInformation",
]



import attr
import structlog

from dlms_cosem.hdlc import address, exceptions, frames
from dlms_cosem.hdlc.exceptions import LocalProtocolError
from dlms_cosem.hdlc.state import (
    AWAITING_CONNECTION,
    AWAITING_DISCONNECT,
    AWAITING_RESPONSE,
    NEED_DATA,
    HdlcConnectionState,
)

LOG = structlog.get_logger()


@attr.s(auto_attribs=True)
class HdlcConnection:
    """
    HDLC - High-level Data Link Control

    In DLMS/COSEM HDLC is used to send DLMS data over serial interfaces. Like the optical
    probe for diagnostic. Some meters also send data over TCP sockets sing HDLC when they
    have not implemented the TCP transport variant of DLMS/COSEM.

    Tracks the state of HDLC communication and transforming bytes to frames and
    frames to bytes.
    """

    client_address: address.HdlcAddress
    server_address: address.HdlcAddress
    client_ssn: int = attr.ib(init=False, default=0)
    client_rsn: int = attr.ib(init=False, default=0)
    server_ssn: int = attr.ib(init=False, default=0)
    server_rsn: int = attr.ib(init=False, default=0)
    max_data_size: int = attr.ib(default=128)
    state: HdlcConnectionState = attr.ib(factory=HdlcConnectionState)
    buffer: bytearray = attr.ib(factory=bytearray)
    buffer_search_position: int = 1

    def send(self, frame) -> bytes:
        """
        Returns the bytes to be sent over I/O for a frame and changes the connection
        state depending on frame.
        :param frame: HDLC frame:
        :return: bytes
        """
        self.state.process_frame(frame)

        if isinstance(frame, frames.InformationFrame):
            self.handle_sequence_numbers(
                frame_ssn=frame.send_sequence_number,
                frame_rsn=frame.receive_sequence_number,
                response=False,
            )

        return frame.to_bytes()

    def handle_sequence_numbers(self, frame_ssn: int, frame_rsn: int, response: bool):
        if not response:
            if frame_ssn != self.server_ssn or frame_rsn != self.server_rsn:
                raise LocalProtocolError(
                    f"Frame sequence numbers are wrong: frame(ssn: {frame_ssn}, rsn: "
                    f"{frame_rsn}) =! client(ssn:{self.server_ssn}, "
                    f"rsn:{self.server_rsn})"
                )
            self.server_ssn += 1
            self.client_rsn += 1
        else:
            if frame_ssn != self.client_ssn or frame_rsn != self.client_rsn:
                raise LocalProtocolError(
                    f"Frame sequence numbers are wrong: frame(ssn: {frame_ssn}, rsn: "
                    f"{frame_rsn}) =! client(ssn:{self.server_ssn}, "
                    f"rsn:{self.server_rsn})"
                )
            self.server_rsn += 1
            self.client_ssn += 1

        if self.server_rsn > 7:
            self.server_rsn = 0
        if self.server_ssn > 7:
            self.server_ssn = 0
        if self.client_rsn > 7:
            self.client_rsn = 0
        if self.client_ssn > 7:
            self.client_ssn = 0

    def receive_data(self, data: bytes):
        """
        Add data into the receive buffer.
        After this you could call next_event
        """
        if data:
            LOG.debug(f"Added data to buffer", data=data)
            self.buffer += data

    def next_event(self):
        """
        Will try to parse a frame from the buffer. If a frame is found the buffer is
        cleared of the bytes making up the frame and the frame is returned.
        If the frame is not parsable we assume it is not compleate and we return a
        NEED_DATA event to signal we need to receive more data.
        :return:
        """
        frame_bytes = self._find_frame()
        if frame_bytes is None:
            return NEED_DATA

        if (
            self.state.current_state == AWAITING_CONNECTION
            or self.state.current_state == AWAITING_DISCONNECT
        ):

            try:
                frame = frames.UnNumberedAcknowledgmentFrame.from_bytes(frame_bytes)
            except (exceptions.HdlcParsingError, ValueError):
                frame = None

        elif self.state.current_state == AWAITING_RESPONSE:
            # It can be a InformationFrame or a ReceiveReadyFrame in case we have sent
            # a segmented frame.
            try:
                frame = frames.InformationFrame.from_bytes(frame_bytes)
            except (exceptions.HdlcParsingError, ValueError):
                # Not an information frame. Should be a receive ready frame
                frame = None

            if frame is None:
                try:
                    frame = frames.ReceiveReadyFrame.from_bytes(frame_bytes)
                except (exceptions.HdlcParsingError, ValueError):
                    frame = None
        else:
            frame = None

        if frame is None:
            LOG.debug("HDLC frame could not be parsed. Need more data")
            return NEED_DATA

        LOG.debug(f"Received HDLC frame", frame=frame)
        self.state.process_frame(frame)
        self._tidy_buffer()

        if isinstance(frame, frames.InformationFrame):
            self.handle_sequence_numbers(
                frame_ssn=frame.send_sequence_number,
                frame_rsn=frame.receive_sequence_number,
                response=True,
            )

        return frame

    def _find_frame(self):
        """
        To find a frame in the buffer we need to assume somethings.
        1. The first character in the buffer should be the HDLC_FLAG.
            During normal operations we will have frames with flags on both ends.
            But with windowing one might be omitted in long information frame exchanges
            Ex: 7e{frame}7e{frame}7e. The second one would not have an initial 7e after
             we take out the first frame.
            So if the intial byte is not 7e we should manually add it.

        2. We might find an incomplete frame if the second 7e was found as data and not
            actually an end flag. So we need to keep the current end memory so we can
            extend the search if we cant parse the frame.

        3. Once we have parsed a proper frame we shoudl call clear buffer that will
        remove the data for the frame from the buffer.
        :return:
        """
        try:
            frame_end = (
                self.buffer.index(frames.HDLC_FLAG, self.buffer_search_position) + 1
            )
        except ValueError:
            # .index raises ValueError on not finding subsection
            return None

        frame_bytes = self.buffer[:frame_end]
        self.buffer_search_position = frame_end

        if not frame_bytes.startswith(frames.HDLC_FLAG):
            frame_bytes.insert(0, ord(frames.HDLC_FLAG))

        return frame_bytes

    def _tidy_buffer(self):
        """
        Remove the bytes we have extracted.
        """
        del self.buffer[: self.buffer_search_position]
        self.buffer_search_position = 1



from cryptography.hazmat.primitives.ciphers import algorithms, modes
from cryptography.hazmat.primitives.ciphers.base import Cipher

from dlms_cosem.security import SecurityControlField, decrypt, encrypt, gmac


def test_encrypt():
    key = b"SUCHINSECUREKIND"
    auth_key = key

    text = b"SUPER_SECRET_TEXT"

    ctext = encrypt(
        key=key,
        auth_key=auth_key,
        invocation_counter=1,
        security_control=SecurityControlField(
            security_suite=0, authenticated=True, encrypted=True
        ),
        system_title=b"12345678",
        plain_text=text,
    )

    print(ctext)

    out = decrypt(
        key=key,
        auth_key=auth_key,
        invocation_counter=1,
        security_control=SecurityControlField(
            security_suite=0, authenticated=True, encrypted=True
        ),
        system_title=b"12345678",
        cipher_text=ctext,
    )

    assert text == out


def test_encrypt_authenticated():
    security_control = SecurityControlField(
        security_suite=0, authenticated=True, encrypted=True
    )
    encryption_key = bytes.fromhex("000102030405060708090A0B0C0D0E0F")
    authentication_key = bytes.fromhex("D0D1D2D3D4D5D6D7D8D9DADBDCDDDEDF")
    system_title = bytes.fromhex("4D4D4D0000BC614E")
    invocation_counter = int.from_bytes(bytes.fromhex("01234567"), "big")
    # Get request attr 2 of clock object.
    plain_data = bytes.fromhex("C0010000080000010000FF0200")

    ciphered_text = bytes.fromhex("411312FF935A47566827C467BC7D825C3BE4A77C3FCC056B6B")

    assert (
        encrypt(
            security_control=security_control,
            key=encryption_key,
            auth_key=authentication_key,
            system_title=system_title,
            invocation_counter=invocation_counter,
            plain_text=plain_data,
        )
        == ciphered_text
    )


def test_decrypt_authenticated():
    security_control = SecurityControlField(
        security_suite=0, authenticated=True, encrypted=True
    )
    encryption_key = bytes.fromhex("000102030405060708090A0B0C0D0E0F")
    authentication_key = bytes.fromhex("D0D1D2D3D4D5D6D7D8D9DADBDCDDDEDF")
    system_title = bytes.fromhex("4D4D4D0000BC614E")
    invocation_counter = int.from_bytes(bytes.fromhex("01234567"), "big")
    # Get request attr 2 of clock object.
    plain_data = bytes.fromhex("C0010000080000010000FF0200")

    ciphered_text = bytes.fromhex("411312FF935A47566827C467BC7D825C3BE4A77C3FCC056B6B")

    assert (
        decrypt(
            security_control=security_control,
            key=encryption_key,
            auth_key=authentication_key,
            system_title=system_title,
            invocation_counter=invocation_counter,
            cipher_text=ciphered_text,
        )
        == plain_data
    )


def test_gmac():
    encryption_key = bytes.fromhex("000102030405060708090A0B0C0D0E0F")
    authentication_key = bytes.fromhex("D0D1D2D3D4D5D6D7D8D9DADBDCDDDEDF")
    security_control = SecurityControlField(
        security_suite=0, authenticated=True, encrypted=False
    )
    client_invocation_counter = int.from_bytes(bytes.fromhex("00000001"), "big")
    client_system_title = bytes.fromhex("4D4D4D0000000001")
    # server_system_title = bytes.fromhex("4D4D4D0000BC614E")
    # server_invocation_counter = int.from_bytes(bytes.fromhex("01234567"), "big")
    # client_to_server_challenge = bytes.fromhex("4B35366956616759")
    server_to_client_challenge = bytes.fromhex("503677524A323146")
    result = gmac(
        security_control=security_control,
        key=encryption_key,
        auth_key=authentication_key,
        invocation_counter=client_invocation_counter,
        system_title=client_system_title,
        challenge=server_to_client_challenge,
    )
    assert len(result) == 12
    assert result == bytes.fromhex("1A52FE7DD3E72748973C1E28")


def test_gmac2():
    encryption_key = bytes.fromhex("000102030405060708090A0B0C0D0E0F")
    authentication_key = bytes.fromhex("D0D1D2D3D4D5D6D7D8D9DADBDCDDDEDF")
    security_control = SecurityControlField(
        security_suite=0, authenticated=True, encrypted=False
    )
    client_invocation_counter = int.from_bytes(bytes.fromhex("00000001"), "big")
    client_system_title = bytes.fromhex("4D4D4D0000000001")
    # server_system_title = bytes.fromhex("4D4D4D0000BC614E")
    # server_invocation_counter = int.from_bytes(bytes.fromhex("01234567"), "big")
    # client_to_server_challenge = bytes.fromhex("4B35366956616759")
    server_to_client_challenge = bytes.fromhex("503677524A323146")

    iv = client_system_title + client_invocation_counter.to_bytes(4, "big")

    assert iv == bytes.fromhex("4D4D4D000000000100000001")

    # Construct an AES-GCM Cipher object with the given key and iv
    encryptor = Cipher(
        algorithms.AES(encryption_key),
        modes.GCM(initialization_vector=iv, tag=None, min_tag_length=12),
    ).encryptor()

    # associated_data will be authenticated but not encrypted,
    # it must also be passed in on decryption.
    associated_data = (
        security_control.to_bytes() + authentication_key + server_to_client_challenge
    )

    assert associated_data == bytes.fromhex(
        "10D0D1D2D3D4D5D6D7D8D9DADBDCDDDEDF503677524A323146"
    )
    encryptor.authenticate_additional_data(associated_data)

    # Encrypt the plaintext and get the associated ciphertext.
    # GCM does not require padding.
    ciphertext = encryptor.update(b"") + encryptor.finalize()

    # dlms uses a tag lenght of 12 not the default of 16. Since we have set the minimum
    # tag length to 12 it is ok to truncated the tag.
    tag = encryptor.tag[:12]

    assert ciphertext == b""
    result = ciphertext + tag

    assert result == bytes.fromhex("1A52FE7DD3E72748973C1E28")



from enum import IntEnum
from typing import *

import attr

from dlms_cosem import enumerations
from dlms_cosem.a_xdr import Attribute, AXdrDecoder, Choice, EncodingConf
from dlms_cosem.protocol.xdlms.base import AbstractXDlmsApdu


class ErrorFactory:

    ERROR_TYPE_MAP: ClassVar[Dict[int, Type[IntEnum]]] = {
        0: enumerations.ApplicationReferenceError,
        1: enumerations.HardwareResourceError,
        2: enumerations.VdeStateError,
        3: enumerations.ServiceError,
        4: enumerations.DefinitionError,
        5: enumerations.AccessError,
        6: enumerations.InitiateError,
        7: enumerations.LoadDataError,
        8: enumerations.DataScopeError,
        9: enumerations.DataScopeError,
        10: enumerations.OtherError,
    }

    @classmethod
    def get_error_type(cls, tag: int):
        return cls.ERROR_TYPE_MAP[tag]


def make_error(source_bytes: bytes):
    if len(source_bytes) != 2:
        raise ValueError(f"Length needs to be 2 not {len(source_bytes)}")
    error_tag = source_bytes[0]
    error_type = ErrorFactory.get_error_type(error_tag)
    return error_type(source_bytes[1])


@attr.s(auto_attribs=True)
class ConfirmedServiceError(AbstractXDlmsApdu):

    TAG: ClassVar[int] = 14

    ENCODING_CONF: ClassVar[EncodingConf] = EncodingConf(
        attributes=[
            Choice(
                choices={
                    b"\x01": Attribute(
                        attribute_name="error", create_instance=make_error, length=2
                    ),
                    b"\x05": Attribute(
                        attribute_name="error", create_instance=make_error, length=2
                    ),
                    b"\x06": Attribute(
                        attribute_name="error", create_instance=make_error, length=2
                    ),
                }
            )
        ]
    )

    error: IntEnum

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(
                f"Tag for ConformedServiceError should be {cls.TAG} not {tag}"
            )
        decoder = AXdrDecoder(cls.ENCODING_CONF)
        result = decoder.decode(data)

        return cls(**result)

    def to_bytes(self) -> bytes:
        # TODO: No good handling of reversing choice in A-XDR. Just setting it
        #  to 01 InitiateError

        rev_error_map = {y: x for x, y in ErrorFactory.ERROR_TYPE_MAP.items()}
        error_type_id = rev_error_map[type(self.error)]

        return bytes([self.TAG, 1, error_type_id, self.error.value])
